# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import re
import subprocess
import sys
import webbrowser
from itertools import cycle
from operator import itemgetter

from trackma import messenger
from trackma import utils
from trackma.accounts import AccountManager
from trackma.engine import Engine

try:
    import urwid
except ImportError:
    print("urwid not found. Make sure you installed the "
          "urwid package.")
    sys.exit(-1)


class Trackma_urwid:
    """
    Main class for the urwid version of Trackma
    """

    """Main objects"""
    engine = None
    mainloop = None
    cur_sort = 'title'
    sorts_iter = cycle(('my_progress', 'total', 'my_score', 'id', 'title'))
    cur_order = False
    orders_iter = cycle((True, False))
    keymapping = dict()
    positions = list()
    last_search = None
    last_update_prompt = ()

    """Widgets"""
    header = None
    listbox = None
    view = None

    def __init__(self):
        """Creates main widgets and creates mainloop"""
        self.config = utils.parse_config(utils.to_config_path(
            'ui-curses.json'), utils.curses_defaults)
        keymap = utils.curses_defaults['keymap']
        keymap.update(self.config['keymap'])
        self.keymap_str = self.get_keymap_str(keymap)
        self.keymapping = self.map_key_to_func(keymap)

        palette = []
        for k, color in self.config['palette'].items():
            palette.append((k, *color))

        # Prepare header
        sys.stdout.write("\x1b]0;Trackma-curses "+utils.VERSION+"\x07")
        self.header_title = urwid.Text('Trackma-curses ' + utils.VERSION)
        self.header_api = urwid.Text('API:')
        self.header_filter = urwid.Text('Filter:')
        self.header_sort = urwid.Text('Sort:title')
        self.header_order = urwid.Text('Order:d')
        self.header = urwid.AttrMap(urwid.Columns([
            self.header_title,
            ('fixed', 30, self.header_filter),
            ('fixed', 17, self.header_sort),
            ('fixed', 16, self.header_api)]), 'status')

        top_pile = [self.header]

        if self.config['show_help']:
            top_text = "{help}:Help  {sort}:Sort  " + \
                       "{update}:Update  {play}:Play  " + \
                       "{status}:Status  {score}:Score  " + \
                       "{quit}:Quit"
            top_text = top_text.format(**self.keymap_str)
            top_pile.append(urwid.AttrMap(urwid.Text(top_text), 'status'))

        self.top_pile = urwid.Pile(top_pile)

        # Prepare status bar
        self.status_text = urwid.Text('Trackma-curses '+utils.VERSION)
        self.status_queue = urwid.Text('Q:N/A')
        self.status_tracker = urwid.Text('T:N/A')
        self.statusbar = urwid.AttrMap(urwid.Columns([
            self.status_text,
            ('fixed', 10, self.status_tracker),
            ('fixed', 6, self.status_queue),
        ]), 'status')

        self.listheader = urwid.AttrMap(
            urwid.Columns([
                ('weight', 1, urwid.Text('Title')),
                ('fixed', 10, urwid.Text('Progress')),
                ('fixed', 7, urwid.Text('Score')),
            ]), 'header')

        self.listwalker = ShowWalker([])
        self.listbox = urwid.ListBox(self.listwalker)
        self.listframe = urwid.Frame(self.listbox, header=self.listheader)

        self.viewing_info = False

        self.view = urwid.Frame(
            self.listframe, header=self.top_pile, footer=self.statusbar)
        self.mainloop = urwid.MainLoop(
            self.view, palette, unhandled_input=self.keystroke, screen=urwid.raw_display.Screen())
        self.mainloop.screen.set_terminal_properties(colors=256)

    def run(self):
        self.mainloop.set_alarm_in(0, self.do_switch_account)
        self.mainloop.run()

    def map_key_to_func(self, keymap):
        keymapping = dict()
        funcmap = {'help': self.do_help,
                   'prev_filter': self.do_prev_filter,
                   'next_filter': self.do_next_filter,
                   'sort': self.do_sort,
                   'sort_order': self.change_sort_order,
                   'update': self.do_update,
                   'play': self.do_play,
                   'openfolder': self.do_openfolder,
                   'play_random': self.do_play_random,
                   'status': self.do_status,
                   'score': self.do_score,
                   'send': self.do_send,
                   'retrieve': self.do_retrieve,
                   'addsearch': self.do_addsearch,
                   'reload': self.do_reload,
                   'switch_account': self.do_switch_account,
                   'delete': self.do_delete,
                   'quit': self.do_quit,
                   'altname': self.do_altname,
                   'search': self.do_search,
                   'neweps': self.do_neweps,
                   'details': self.do_info,
                   'details_exit': self.do_info_exit,
                   'open_web': self.do_open_web,
                   'left': self.key_left,
                   'down': self.key_down,
                   'up': self.key_up,
                   'right': self.key_right,
                   'page_down': self.key_page_down,
                   'page_up': self.key_page_up,
                   }

        for func, keybind in keymap.items():
            try:
                if isinstance(keybind, list):
                    for keybindm in keybind:
                        keymapping[keybindm] = funcmap[func]
                else:
                    keymapping[keybind] = funcmap[func]
            except KeyError:
                # keymap.json requested an action not available in funcmap
                pass
        return keymapping

    def get_keymap_str(self, keymap):
        stringed = {}
        for k, keybind in keymap.items():
            if isinstance(keybind, list):
                stringed[k] = ','.join(keybind)
            else:
                stringed[k] = keybind
        return stringed

    def _rebuild(self):
        self.header_api.set_text('API:%s' % self.engine.api_info['name'])
        self.lists = dict()
        self.filters = self.engine.mediainfo['statuses_dict']
        self.filters_nums = self.engine.mediainfo['statuses']
        self.filters_sizes = []

        track_info = self.engine.tracker_status()
        if track_info:
            self.tracker_state(track_info)

        for status in self.filters_nums:
            self.lists[status] = urwid.ListBox(ShowWalker([]))

        self._rebuild_lists()

        # Put the number of shows in every status in a list
        for status in self.filters_nums:
            self.filters_sizes.append(len(self.lists[status].body))

        self.set_filter(0)
        self.status('Ready.')
        self.started = True

    def _rebuild_lists(self, status=None):
        if status:
            self.lists[status].body[:] = []
            showlist = self.engine.filter_list(status)
        else:
            for _status in self.lists.keys():
                self.lists[_status].body[:] = []
            showlist = self.engine.get_list()

        library = self.engine.library()
        sortedlist = sorted(showlist, key=itemgetter(
            self.cur_sort), reverse=self.cur_order)

        for show in sortedlist:
            item = ShowItem(show, self.engine.mediainfo['has_progress'], self.engine.altname(
                show['id']), library.get(show['id']))

            self.lists[show['my_status']].body.append(item)

    def start(self, account):
        """Starts the engine"""
        # Engine configuration
        self.started = False

        self.status("Starting engine...")
        self.engine = Engine(account, self.message_handler)
        self.engine.connect_signal('episode_changed', self.changed_show)
        self.engine.connect_signal('score_changed', self.changed_show)
        self.engine.connect_signal('status_changed', self.changed_show_status)
        self.engine.connect_signal('playing', self.playing_show)
        self.engine.connect_signal('show_added', self.changed_list)
        self.engine.connect_signal('show_deleted', self.changed_list)
        self.engine.connect_signal('show_synced', self.changed_show)
        self.engine.connect_signal('queue_changed', self.changed_queue)
        self.engine.connect_signal('prompt_for_update', self.prompt_update)
        self.engine.connect_signal('tracker_state', self.tracker_state)

        # Engine start and list rebuild
        self.status("Building lists...")
        self.engine.start()
        self._rebuild()

    def set_filter(self, filter_num):
        self.cur_filter = filter_num
        _filter = self.filters_nums[self.cur_filter]
        self.header_filter.set_text("Filter:%s (%d)" % (
            self.filters[_filter], self.filters_sizes[self.cur_filter]))

        self.listframe.body = self.lists[_filter]

    def _get_cur_list(self):
        _filter = self.filters_nums[self.cur_filter]
        return self.lists[_filter].body

    def _get_selected_item(self):
        return self._get_cur_list().get_focus()[0]

    def status(self, msg):
        self.status_text.set_text(msg)

    def error(self, msg):
        self.status_text.set_text([('error', "Error: %s" % msg)])

    def message_handler(self, classname, msgtype, msg):
        if msgtype != messenger.TYPE_DEBUG:
            try:
                self.status(msg)
                self.mainloop.draw_screen()
            except AssertionError:
                print(msg)

    def keystroke(self, key_input):
        try:
            self.keymapping[key_input]()
        except KeyError:
            # Unbound key pressed; do nothing
            pass

    def key_left(self):
        self.mainloop.process_input(['left'])

    def key_down(self):
        self.mainloop.process_input(['down'])

    def key_up(self):
        self.mainloop.process_input(['up'])

    def key_right(self):
        self.mainloop.process_input(['right'])

    def key_page_down(self):
        self.mainloop.process_input(['page down'])

    def key_page_up(self):
        self.mainloop.process_input(['page up'])

    def forget_account(self):
        manager = AccountManager()
        manager.set_default(None)

    def do_switch_account(self, loop=None, data=None):
        manager = AccountManager()

        if self.engine is None:
            if manager.get_default():
                self.start(manager.get_default())
            else:
                self.dialog = AccountDialog(self.mainloop, manager, False)
                urwid.connect_signal(self.dialog, 'done', self.start)
        else:
            self.dialog = AccountDialog(self.mainloop, manager, True)
            urwid.connect_signal(self.dialog, 'done', self.do_reload_engine)

    def do_addsearch(self):
        self.ask('Search on remote: ', self.addsearch_request)

    def do_delete(self):
        if self._get_selected_item():
            self.question('Delete selected show? [y/n] ', self.delete_request)

    def do_prev_filter(self):
        if self.cur_filter > 0:
            self.set_filter(self.cur_filter - 1)

    def do_next_filter(self):
        if self.cur_filter < len(self.filters)-1:
            self.set_filter(self.cur_filter + 1)

    def do_sort(self):
        self.status("Sorting...")
        _sort = next(self.sorts_iter)
        self.cur_sort = _sort
        self.header_sort.set_text("Sort:%s" % _sort)
        self._rebuild_lists()
        self.status("Ready.")

    def change_sort_order(self):
        self.status("Sorting...")
        _order = next(self.orders_iter)
        self.cur_order = _order
        self._rebuild_lists()
        self.status("Ready.")

    def do_update(self):
        item = self._get_selected_item()
        if item:
            show = self.engine.get_show_info(item.showid)
            self.ask('[Update] Episode # to update to: ',
                     self.update_request, show['my_progress']+1)

    def do_play(self):
        item = self._get_selected_item()
        if item:
            show = self.engine.get_show_info(item.showid)
            self.ask('[Play] Episode # to play: ',
                     self.play_request, show['my_progress']+1)

    def do_openfolder(self):
        item = self._get_selected_item()
        try:
            self.engine.open_show_folder(item.showid)
        except utils.EngineError as e:
            self.error(e.args[0])

    def do_play_random(self):
        try:
            args = self.engine.play_random()
            utils.spawn_process(args)
        except utils.TrackmaError as e:
            self.error(e)
            return

    def do_send(self):
        self.engine.list_upload()
        self.status("Ready.")

    def do_retrieve(self):
        try:
            self.engine.list_download()
            self._rebuild_lists()
            self.status("Ready.")
        except utils.TrackmaError as e:
            self.error(e)

    def do_help(self):
        helptext = "Trackma-curses " + utils.VERSION + \
            "  by z411 (z411@omaera.org)\n\n"
        helptext += "Trackma is an open source client for media tracking websites.\n"
        helptext += "https://github.com/z411/trackma\n\n"
        helptext += "This program is licensed under the GPLv3,\nfor more information read COPYING file.\n\n"
        helptext += "More controls:\n  {prev_filter}/{next_filter}:Change Filter\n  {search}:Search\n  {addsearch}:Add\n  {reload}:Change API/Mediatype\n"
        helptext += "  {delete}:Delete\n  {send}:Send changes\n  {sort_order}:Change sort order\n  {retrieve}:Retrieve list\n  {details}: View details\n  {open_web}: Open website\n  {openfolder}: Open folder containing show\n  {altname}:Set alternative title\n  {neweps}:Search for new episodes\n  {play_random}:Play Random\n  {switch_account}: Change account"
        helptext = helptext.format(**self.keymap_str)
        ok_button = urwid.Button('OK', self.help_close)
        ok_button_wrap = urwid.Padding(urwid.AttrMap(
            ok_button, 'button', 'button hilight'), 'center', 6)
        pile = urwid.Pile([urwid.Text(helptext), ok_button_wrap])
        self.dialog = Dialog(pile, self.mainloop, width=62, title='About/Help')
        self.dialog.show()

    def help_close(self, widget):
        self.dialog.close()

    def do_altname(self):
        item = self._get_selected_item()
        if item:
            show = self.engine.get_show_info(item.showid)
            self.status(show['title'])
            self.ask('[Altname] New alternative name: ',
                     self.altname_request, self.engine.altname(item.showid))

    def do_score(self):
        item = self._get_selected_item()
        if item:
            show = self.engine.get_show_info(item.showid)
            self.ask('[Score] Score to change to: ',
                     self.score_request, show['my_score'])

    def do_status(self):
        item = self._get_selected_item()
        if not item:
            return

        show = self.engine.get_show_info(item.showid)

        buttons = list()
        num = 1
        selected = 1
        title = urwid.Text('Choose status:')
        title.align = 'center'
        buttons.append(title)
        for status in self.filters_nums:
            name = self.filters[status]
            button = urwid.Button(name, self.status_request, status)
            button._label.align = 'center'
            buttons.append(urwid.AttrMap(button, 'button', 'button hilight'))
            if status == show['my_status']:
                selected = num
            num += 1
        pile = urwid.Pile(buttons)
        pile.set_focus(selected)
        self.dialog = Dialog(pile, self.mainloop, width=22)
        self.dialog.show()

    def do_reload(self):
        # Create a list of buttons to select the mediatype
        rb_mt = []
        mediatypes = []
        for mediatype in self.engine.api_info['supported_mediatypes']:
            but = urwid.RadioButton(rb_mt, mediatype)
            # Make it selected if it's the current mediatype
            if self.engine.api_info['mediatype'] == mediatype:
                but.set_state(True)
            urwid.connect_signal(
                but, 'change', self.reload_request, [None, mediatype])
            mediatypes.append(urwid.AttrMap(but, 'button', 'button hilight'))
        mediatype = urwid.Columns(
            [urwid.Text('Mediatype:'), urwid.Pile(mediatypes)])

        # main_pile = urwid.Pile([mediatype, urwid.Divider(), api])
        self.dialog = Dialog(mediatype, self.mainloop,
                             width=30, title='Change media type')
        self.dialog.show()

    def do_reload_engine(self, account=None, mediatype=None):
        self.started = False
        self.engine.reload(account, mediatype)
        self._rebuild()

    def do_open_web(self):
        item = self._get_selected_item()
        if item:
            show = self.engine.get_show_info(item.showid)
            if show['url']:
                webbrowser.open(show['url'], 2, True)

    def do_info(self):
        if self.viewing_info:
            return

        item = self._get_selected_item()
        if not item:
            return

        show = self.engine.get_show_info(item.showid)

        self.status("Getting show details...")

        try:
            details = self.engine.get_show_details(show)
        except utils.TrackmaError as e:
            self.error(e)
            return

        title = urwid.Text(('info_title', show['title']), 'center', 'any')
        widgets = []
        for line in details['extra']:
            if line[0] and line[1]:
                widgets.append(urwid.Text(('info_section', "%s: " % line[0])))
                if isinstance(line[1], dict):
                    linestr = repr(line[1])
                elif isinstance(line[1], int) or isinstance(line[1], list):
                    linestr = str(line[1])
                else:
                    linestr = line[1]

                widgets.append(urwid.Padding(
                    urwid.Text("{}\n".format(linestr)), left=3))

        self.view.body = urwid.Frame(urwid.ListBox(widgets), header=title)
        self.viewing_info = True
        self.status("Detail View | ESC:Return  Up/Down:Scroll  O:View website")

    def do_info_exit(self):
        if self.viewing_info:
            self.view.body = self.listframe
            self.viewing_info = False
            self.status("Ready.")

    def do_neweps(self):
        try:
            shows = self.engine.scan_library(rescan=True)
            self._rebuild_lists()

            self.status("Ready.")
        except utils.TrackmaError as e:
            self.error(e)

    def do_quit(self):
        if self.viewing_info:
            self.do_info_exit()
        else:
            self.engine.unload()
            raise urwid.ExitMainLoop()

    def addsearch_request(self, data):
        self.ask_finish(self.addsearch_request)
        if data:
            try:
                shows = self.engine.search(data)
            except utils.TrackmaError as e:
                self.error(e)
                return

            if len(shows) > 0:
                self.status("Ready.")
                self.dialog = AddDialog(
                    self.mainloop, self.engine, showlist=shows, width=('relative', 80))
                urwid.connect_signal(self.dialog, 'done', self.addsearch_do)
                self.dialog.show()
            else:
                self.status("No results.")

    def addsearch_do(self, show):
        self.dialog.close()
        # Add show as current status
        _filter = self.filters_nums[self.cur_filter]
        try:
            self.engine.add_show(show, _filter)
        except utils.TrackmaError as e:
            self.error(e)

    def delete_request(self, data):
        self.ask_finish(self.delete_request)
        if data == 'y':
            showid = self._get_selected_item().showid
            show = self.engine.get_show_info(showid)

            try:
                self.engine.delete_show(show)
            except utils.TrackmaError as e:
                self.error(e)

    def status_request(self, widget, data=None):
        self.dialog.close()
        if data is not None:
            item = self._get_selected_item()

            try:
                show = self.engine.set_status(item.showid, data)
            except utils.TrackmaError as e:
                self.error(e)
                return

    def reload_request(self, widget, selected, data):
        if selected:
            self.dialog.close()
            self.do_reload_engine(data[0], data[1])

    def update_request(self, data):
        self.ask_finish(self.update_request)
        if data:
            item = self._get_selected_item()

            try:
                show = self.engine.set_episode(item.showid, data)
            except utils.TrackmaError as e:
                self.error(e)
                return

    def score_request(self, data):
        self.ask_finish(self.score_request)
        if data:
            item = self._get_selected_item()

            try:
                show = self.engine.set_score(item.showid, data)
            except utils.TrackmaError as e:
                self.error(e)
                return

    def altname_request(self, data):
        self.ask_finish(self.altname_request)
        if data:
            item = self._get_selected_item()

            try:
                self.engine.altname(item.showid, data)
                item.update_altname(self.engine.altname(item.showid))
            except utils.TrackmaError as e:
                self.error(e)
                return

    def play_request(self, data):
        self.ask_finish(self.play_request)
        if data:
            item = self._get_selected_item()
            show = self.engine.get_show_info(item.showid)

            try:
                args = self.engine.play_episode(show, data)
                utils.spawn_process(args)
            except utils.TrackmaError as e:
                self.error(e)
                return

    def prompt_update_request(self, data):
        (show, episode) = self.last_update_prompt
        self.ask_finish(self.prompt_update_request)
        if data == 'y':
            try:
                show = self.engine.set_episode(show['id'], episode)
            except utils.TrackmaError as e:
                self.error(e)
                return
        else:
            self.status('Ready.')

    def prompt_update(self, show, episode):
        self.last_update_prompt = (show, episode)
        self.question("Update %s to episode %d? [y/N] " %
                      (show['title'], episode), self.prompt_update_request)

    def changed_show(self, show, changes=None):
        if self.started and show:
            status = show['my_status']
            self.lists[status].body.update_show(show)
            self.mainloop.draw_screen()

    def changed_show_status(self, show, old_status=None):
        self._rebuild_lists(show['my_status'])
        if old_status is not None:
            self._rebuild_lists(old_status)

        go_filter = 0
        for _filter in self.filters_nums:
            if _filter == show['my_status']:
                break
            go_filter += 1

        self.set_filter(go_filter)
        self._get_cur_list().select_show(show)

    def changed_queue(self, queue):
        self.status_queue.set_text("Q:{}".format(len(queue)))

    def tracker_state(self, status):
        state = status['state']
        timer = status['timer']
        paused = status['paused']

        if state == utils.Tracker.NOVIDEO:
            st = 'LISTEN'
        elif state == utils.Tracker.PLAYING:
            st = '{}{}'.format('#' if paused else '+', timer)
        elif state == utils.Tracker.UNRECOGNIZED:
            st = 'UNRECOG'
        elif state == utils.Tracker.NOT_FOUND:
            st = 'NOTFOUN'
        elif state == utils.Tracker.IGNORED:
            st = 'IGNORE'
        else:
            st = '???'

        self.status_tracker.set_text("T:{}".format(st))
        self.mainloop.draw_screen()

    def playing_show(self, show, is_playing, episode=None):
        status = show['my_status']
        self.lists[status].body.playing_show(show, is_playing)
        self.mainloop.draw_screen()

    def changed_list(self, show):
        self._rebuild_lists(show['my_status'])

    def ask(self, msg, callback, data=u''):
        self.asker = Asker(msg, str(data))
        self.view.set_footer(urwid.AttrMap(self.asker, 'status'))
        self.view.set_focus('footer')
        urwid.connect_signal(self.asker, 'done', callback)

    def question(self, msg, callback, data=u''):
        self.asker = QuestionAsker(msg, str(data))
        self.view.set_footer(urwid.AttrMap(self.asker, 'status'))
        self.view.set_focus('footer')
        urwid.connect_signal(self.asker, 'done', callback)

    def ask_finish(self, callback):
        self.view.set_focus('body')
        urwid.disconnect_signal(self, self.asker, 'done', callback)
        self.view.set_footer(self.statusbar)

    def do_search(self, key=''):
        if self.last_search:
            text = "Search forward [%s]: " % self.last_search
        else:
            text = "Search forward: "

        self.ask(text, self.search_request, key)
        # urwid.connect_signal(self.asker, 'change', self.search_live)

    # def search_live(self, widget, data):
    #    if data:
    #        self.listwalker.select_match(data)

    def search_request(self, data):
        self.ask_finish(self.search_request)
        try:
            if data:
                self.last_search = data
                self._get_cur_list().select_match(data)
            elif self.last_search:
                self._get_cur_list().select_match(self.last_search)
        except re.error as e:
            self.error(e)


class Dialog(urwid.Overlay):
    def __init__(self, widget, loop, width=30, height=None, title=''):
        self.widget = urwid.AttrMap(
            urwid.LineBox(widget, title=title), 'window')
        self.oldwidget = loop.widget
        self.loop = loop
        super().__init__(self.widget, loop.widget,
                         align="center",
                         width=width,
                         valign="middle",
                         height=height)

    def show(self):
        self.loop.widget = self

    def close(self):
        self.loop.widget = self.oldwidget

    def keypress(self, size, key):
        if key in ('up', 'down', 'left', 'right', 'enter'):
            self.widget.keypress(size, key)
        elif key == 'esc':
            self.close()


class AddDialog(Dialog):
    __metaclass__ = urwid.signals.MetaSignals
    signals = ['done']

    def __init__(self, loop, engine, showlist={}, width=30):
        self.viewing_info = False
        self.engine = engine

        self.listheader = urwid.Columns([
            ('fixed', 7, urwid.Text('ID')),
            ('weight', 1, urwid.Text('Title')),
            ('fixed', 10, urwid.Text('Type')),
            ('fixed', 7, urwid.Text('Total')),
        ])

        self.listwalker = urwid.SimpleListWalker([])
        self.listbox = urwid.ListBox(self.listwalker)

        # Add results to the list
        for show in showlist:
            self.listwalker.append(SearchItem(show))

        self.info_txt = urwid.Text(
            "Add View | Enter:Add  i:Info  O:Website  Esc:Cancel")
        self.frame = urwid.Frame(
            self.listbox, header=self.listheader, footer=self.info_txt)
        super().__init__(self.frame, loop, width=width,
                         height=('relative', 80), title='Search results')

    def keypress(self, size, key):
        if key in ('up', 'down', 'left', 'right', 'tab'):
            self.widget.keypress(size, key)
        elif key == 'enter':
            show = self.listwalker.get_focus()[0].show
            urwid.emit_signal(self, 'done', show)
        elif key == 'i':
            show = self.listwalker.get_focus()[0].show
            self.do_info()
        elif key == 'O':
            show = self.listwalker.get_focus()[0].show
            webbrowser.open(show['url'], 2, True)
        elif key == 'esc':
            self.do_info_exit()

    def do_info(self):
        if self.viewing_info:
            return

        show = self.listwalker.get_focus()[0].show

        self.status("Getting show details...")
        details = self.engine.get_show_details(show)

        title = urwid.Text(('info_title', show['title']), 'center', 'any')
        widgets = []
        for line in details['extra']:
            if line[0] and line[1]:
                widgets.append(urwid.Text(('info_section', "%s: " % line[0])))
                if isinstance(line[1], dict):
                    linestr = repr(line[1])
                elif isinstance(line[1], int):
                    linestr = str(line[1])
                else:
                    linestr = line[1]

                widgets.append(urwid.Padding(
                    urwid.Text(linestr + "\n"), left=3))

        self.frame.body = urwid.ListBox(widgets)
        self.frame.header = title
        self.viewing_info = True
        self.info_txt.set_text(
            "Detail View | ESC:Return  Up/Down:Scroll  O:View website")

    def do_info_exit(self):
        if self.viewing_info:
            self.frame.body = self.listbox
            self.frame.header = self.listheader
            self.info_txt.set_text(
                "Add View | Enter:Add  i:Info  O:Website  Esc:Cancel")
            self.viewing_info = False
        else:
            self.close()


class AccountDialog(Dialog):
    __metaclass__ = urwid.signals.MetaSignals
    signals = ['done']
    adding_data = {'username': None, 'password': None, 'api': None, 'apiname': None, 'extra': {}}

    def __init__(self, loop, manager, switch=False, width=50):
        self.switch = switch
        self.manager = manager

        listheader = urwid.Columns([
            ('weight', 1, urwid.Text('Username')),
            ('fixed', 15, urwid.Text('Site')),
        ])

        self.listwalker = urwid.SimpleListWalker([])
        listbox = urwid.ListBox(self.listwalker)

        self.build_list()

        self.foot = urwid.Text('enter:Use once  r:Use always  a:Add  D:Delete')
        self.frame = urwid.Frame(listbox, header=listheader, footer=self.foot)
        super().__init__(self.frame, loop, width=width, height=15, title='Select Account')

        self.adding = False

        self.show()

    def build_list(self):
        self.listwalker[:] = []

        for k, account in self.manager.get_accounts():
            self.listwalker.append(AccountItem(k, account))

    def keypress(self, size, key):
        # if key in ('up', 'down', 'left', 'right', 'tab'):
        #    self.widget.keypress(size, key)
        if self.adding:
            if key == 'esc':
                self.foot_clear()
            else:
                self.widget.keypress(size, key)
        else:
            if key == 'enter':
                self.do_select(False)
            elif key == 'a':
                self.do_add_api()
            elif key == 'r':
                self.do_select(True)
            elif key == 'D':
                self.do_delete_ask()
            elif key == 'esc':
                self.close()
                if not self.switch:
                    raise urwid.ExitMainLoop()
            else:
                self.widget.keypress(size, key)

    def do_add_api(self):
        self.adding = True
        available_libs = ', '.join(sorted(utils.available_libs.keys()))
        ask = Asker("API (%s): " % available_libs)
        self.frame.footer = ask
        self.frame.set_focus('footer')
        urwid.connect_signal(ask, 'done', self.do_add_username)

    def do_add_username(self, data):
        self.adding_data['apiname'] = data
        try:
            self.adding_data['api'] = api = utils.available_libs[data]
        except KeyError:
            self.adding = False
            self.frame.footer = urwid.Text("Error: Invalid API.")
            self.frame.set_focus('body')
            return

        if api[2] == utils.Login.OAUTH:
            ask = Asker("Account name: ")
        else:
            ask = Asker("Username: ")
        self.frame.footer = ask
        urwid.connect_signal(ask, 'done', self.do_add_password)

    def do_add_password(self, data):
        self.adding_data['username'] = data
        if self.adding_data['api'][2] in [utils.Login.OAUTH, utils.Login.OAUTH_PKCE]:
            auth_url = self.adding_data['api'][3]
            
            if self.adding_data['api'][2] == utils.Login.OAUTH_PKCE:
                self.adding_data['extra'] = {'code_verifier': utils.oauth_generate_pkce()}
                auth_url = auth_url % self.adding_data['extra']['code_verifier']
            
            import webbrowser
            webbrowser.open_new_tab(auth_url)
            
            ask = Asker("Enter PIN: ".format(auth_url))
        else:
            ask = Asker("Password: ")
        self.frame.footer = ask
        urwid.connect_signal(ask, 'done', self.do_add)

    def do_delete_ask(self):
        self.adding = True
        ask = QuestionAsker("Do you want to delete this account? [y/n] ")
        self.frame.footer = ask
        self.frame.set_focus('footer')
        urwid.connect_signal(ask, 'done', self.do_delete)

    def do_delete(self, data):
        if data == 'y':
            accountitem = self.listwalker.get_focus()[0]
            self.manager.delete_account(accountitem.num)

        self.build_list()
        self.foot_clear()

    def do_add(self, data):
        username = self.adding_data['username']
        password = data
        api = self.adding_data['apiname']
        extra = self.adding_data['extra']

        try:
            self.manager.add_account(username, password, api, extra)
        except utils.AccountError as e:
            self.adding = False
            self.frame.footer = urwid.Text("Error: %s" % e)
            self.frame.set_focus('body')
            return

        self.build_list()
        self.foot_clear()

    def foot_clear(self):
        self.adding = False
        self.frame.footer = self.foot
        self.frame.set_focus('body')

    def do_select(self, remember):
        accountitem = self.listwalker.get_focus()[0]
        if remember:
            self.manager.set_default(accountitem.num)
        else:
            self.manager.set_default(None)
        self.close()
        urwid.emit_signal(self, 'done', accountitem.account)


class AccountItem(urwid.WidgetWrap):
    def __init__(self, num, account):
        self.num = num
        self.account = account
        self.item = [
            ('weight', 1, urwid.Text(account['username'])),
            ('fixed', 15, urwid.Text(account['api'])),
        ]
        w = urwid.AttrMap(urwid.Columns(self.item), 'window', 'focus')
        super().__init__(w)

    def selectable(self):
        return True

    def keypress(self, size, key):
        return key


class SearchItem(urwid.WidgetWrap):
    def __init__(self, show, has_progress=True):
        self.show = show
        self.item = [
            ('weight', 1, urwid.Text(show['title'])),
            ('fixed', 10, urwid.Text(str(show['type']))),
            ('fixed', 7, urwid.Text("%d" % show['total'])),
        ]
        w = urwid.AttrMap(urwid.Columns(self.item), 'window', 'focus')
        super().__init__(w)

    def selectable(self):
        return True

    def keypress(self, size, key):
        return key


class ShowWalker(urwid.SimpleListWalker):
    def _get_showitem(self, showid):
        for i, item in enumerate(self):
            if showid == item.showid:
                return (i, item)
        # raise Exception('Show not found in ShowWalker.')
        return (None, None)

    def highlight_show(self, show, tocolor):
        (position, showitem) = self._get_showitem(show['id'])
        showitem.highlight(tocolor)

    def update_show(self, show):
        (position, showitem) = self._get_showitem(show['id'])
        if showitem:
            showitem.update(show)
            return True
        else:
            return False

    def playing_show(self, show, is_playing):
        (position, showitem) = self._get_showitem(show['id'])
        if showitem:
            showitem.playing = is_playing
            showitem.highlight(show)
            return True
        else:
            return False

    def select_show(self, show):
        (position, showitem) = self._get_showitem(show['id'])
        if showitem:
            self.set_focus(position)

    def select_match(self, searchstr):
        pos = self.get_focus()[1]
        for i, item in enumerate(self):
            if i <= pos:
                continue
            if re.search(searchstr, item.showtitle, re.I):
                self.set_focus(i)
                break


class ShowItem(urwid.WidgetWrap):
    def __init__(self, show, has_progress=True, altname=None, eps=None):
        if has_progress:
            self.episodes_str = urwid.Text(
                "{0:3} / {1}".format(show['my_progress'], show['total'] or '?'))
        else:
            self.episodes_str = urwid.Text("-")

        if eps:
            self.eps = eps.keys()
        else:
            self.eps = None

        self.score_str = urwid.Text("{0:^5}".format(show['my_score']))
        self.has_progress = has_progress
        self.playing = False

        self.show = show
        self.showid = show['id']

        self.showtitle = show['title']
        if altname:
            self.showtitle += " (%s)" % altname
        self.title_str = urwid.Text(self.showtitle)

        self.item = [
            ('weight', 1, self.title_str),
            ('fixed', 10, self.episodes_str),
            ('fixed', 7, self.score_str),
        ]

        # If the show should be highlighted, do it
        # otherwise color it according to its status
        if self.playing:
            self.color = 'item_playing'
        elif show.get('queued'):
            self.color = 'item_updated'
        elif self.eps and max(self.eps) > show['my_progress']:
            self.color = 'item_neweps'
        elif show['status'] == utils.Status.AIRING:
            self.color = 'item_airing'
        elif show['status'] == utils.Status.NOTYET:
            self.color = 'item_notaired'
        else:
            self.color = 'body'

        self.m = urwid.AttrMap(urwid.Columns(self.item), self.color, 'focus')

        super().__init__(self.m)

    def get_showid(self):
        return self.showid

    def update(self, show):
        if show['id'] == self.showid:
            # Update progress
            if self.has_progress:
                self.episodes_str.set_text(
                    "{0:3} / {1}".format(show['my_progress'], show['total']))
            self.score_str.set_text("{0:^5}".format(show['my_score']))

            # Update color
            self.highlight(show)
        else:
            print("Warning: Tried to update a show with a different ID! (%d -> %d)" %
                  (show['id'], self.showid))

    def update_altname(self, altname):
        # Update title
        self.showtitle = "%s (%s)" % (self.show['title'], altname)
        self.title_str.set_text(self.showtitle)

    def highlight(self, show):
        if self.playing:
            self.color = 'item_playing'
        elif show.get('queued'):
            self.color = 'item_updated'
        elif self.eps and max(self.eps) > show['my_progress']:
            self.color = 'item_neweps'
        elif show['status'] == 1:
            self.color = 'item_airing'
        elif show['status'] == 3:
            self.color = 'item_notaired'
        else:
            self.color = 'body'

        self.m.set_attr_map({None: self.color})

    def selectable(self):
        return True

    def keypress(self, size, key):
        return key


class Asker(urwid.Edit):
    __metaclass__ = urwid.signals.MetaSignals
    signals = ['done']

    def keypress(self, size, key):
        if key == 'enter':
            urwid.emit_signal(self, 'done', self.get_edit_text())
            return
        elif key == 'esc':
            urwid.emit_signal(self, 'done', None)
            return

        urwid.Edit.keypress(self, size, key)


class QuestionAsker(Asker):
    def keypress(self, size, key):
        if key.lower() in 'yn':
            urwid.emit_signal(self, 'done', key.lower())


def main():
    ui = Trackma_urwid()
    try:
        ui.run()
    except utils.TrackmaFatal as e:
        ui.forget_account()
        print("Fatal error: %s" % e)


if __name__ == '__main__':
    main()



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os

from gi.repository import Gdk, Gtk

from trackma.ui.gtk import gtk_dir
from trackma.ui.gtk.showinfobox import ShowInfoBox


@Gtk.Template.from_file(os.path.join(gtk_dir, 'data/showinfowindow.ui'))
class ShowInfoWindow(Gtk.Dialog):
    __gtype_name__ = 'ShowInfoWindow'

    info_container = Gtk.Template.Child()

    def __init__(self, engine, show_data, transient_for=None):
        Gtk.Dialog.__init__(self, use_header_bar=True,
                            transient_for=transient_for)
        self.init_template()

        self._engine = engine
        self._show = show_data

        info_box = ShowInfoBox(engine)
        info_box.load(show_data)
        info_box.show()

        self.info_container.pack_start(info_box, True, True, 0)

    @Gtk.Template.Callback()
    def _on_dialog_close(self, widget):
        self.destroy()

    @Gtk.Template.Callback()
    def _on_btn_website_clicked(self, btn):
        if self._show['url']:
            Gtk.show_uri(None, self._show['url'], Gdk.CURRENT_TIME)



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import datetime
import gzip
import json
import socket
import time
import urllib.error
import urllib.parse
import urllib.request

from trackma import utils
from trackma.lib.lib import lib


class libmal(lib):
    """
    API class to communicate with MyAnimeList (new API)

    Website: https://myanimelist.net
    """
    name = 'libmal'
    msg = None
    logged_in = False

    api_info = {'name': 'MyAnimeList', 'shortname': 'mal',
                'version': 3, 'merge': False}
    mediatypes = dict()
    mediatypes['anime'] = {
        'has_progress': True,
        'can_add': True,
        'can_delete': True,
        'can_score': True,
        'can_status': True,
        'can_update': True,
        'can_play': True,
        'can_date': True,
        'date_next_ep': True,
        'statuses_start': ['watching'],
        'statuses_finish': ['completed'],
        'statuses_library': ['watching', 'on_hold', 'plan_to_watch'],
        'statuses':  ['watching', 'completed', 'on_hold', 'dropped', 'plan_to_watch'],
        'statuses_dict': {
            'watching': 'Watching',
            'completed': 'Completed',
            'on_hold': 'On hold',
            'dropped': 'Dropped',
            'plan_to_watch': 'Plan to Watch'
        },
        'score_max': 10,
        'score_step': 1,
        'search_methods': [utils.SearchMethod.KW, utils.SearchMethod.SEASON],
    }
    mediatypes['manga'] = {
        'has_progress': True,
        'can_add': True,
        'can_delete': True,
        'can_score': True,
        'can_status': True,
        'can_update': True,
        'can_play': False,
        'can_date': True,
        'statuses_start': ['reading'],
        'statuses_finish': ['completed'],
        'statuses':  ['reading', 'completed', 'on_hold', 'dropped', 'plan_to_read'],
        'statuses_dict': {
            'reading': 'Reading',
            'completed': 'Completed',
            'on_hold': 'On hold',
            'dropped': 'Dropped',
            'plan_to_read': 'Plan to Read'
        },
        'score_max': 10,
        'score_step': 1,
        'search_methods': [utils.SearchMethod.KW],
    }
    default_mediatype = 'anime'

    type_translate = {
        'tv': utils.Type.TV,
        'movie': utils.Type.MOVIE,
        'special': utils.Type.SP,
        'ova': utils.Type.OVA,
        'ona': utils.Type.ONA,
        'music': utils.Type.OTHER,
        'unknown': utils.Type.UNKNOWN,
        'manga': utils.Type.MANGA,
        'novel': utils.Type.NOVEL,
        'light_novel': utils.Type.NOVEL,
        'manhwa': utils.Type.MANGA,
        'manhua': utils.Type.MANGA,
        'one_shot': utils.Type.ONE_SHOT,
        'doujinshi': utils.Type.MANGA,
        'cm':utils.Type.OTHER,
        'tv_special': utils.Type.SPECIAL
    }

    status_translate = {
        'currently_airing': utils.Status.AIRING,
        'finished_airing': utils.Status.FINISHED,
        'not_yet_aired': utils.Status.NOTYET,
    }

    season_translate = {
        utils.Season.WINTER: 'winter',
        utils.Season.SPRING: 'spring',
        utils.Season.SUMMER: 'summer',
        utils.Season.FALL: 'fall',
    }

    # Supported signals for the data handler
    signals = {'show_info_changed': None, }

    auth_url = "https://myanimelist.net/v1/oauth2/token"
    query_url = "https://api.myanimelist.net/v2"
    client_id = "32c510ab2f47a1048a8dd24de266dc0c"
    user_agent = 'Trackma/{}'.format(utils.VERSION)

    library_page_limit = 1000
    search_page_limit = 100
    season_page_limit = 500

    def __init__(self, messenger, account, userconfig):
        super(libmal, self).__init__(messenger, account, userconfig)

        self.pin = account['password'].strip()

        if 'extra' not in account or 'code_verifier' not in account['extra']:
            raise utils.APIFatal(
                "This account seems to be using the old MyAnimeList API."
                "Please re-create and authorize the account.")

        self.code_verifier = account['extra']['code_verifier']
        self.userid = self._get_userconfig('userid')

        if self.mediatype == 'manga':
            self.total_str = "num_chapters"
            self.watched_str = self.watched_send_str = "num_chapters_read"
        else:
            self.total_str = "num_episodes"
            self.watched_str = "num_episodes_watched"
            self.watched_send_str = "num_watched_episodes"  # Please fix this upstream...

        self.opener = urllib.request.build_opener()
        self.opener.addheaders = [
            ('User-Agent',      self.user_agent),
            ('Accept',          'application/json'),
            ('Accept-Encoding', 'gzip'),
            ('Accept-Charset',  'utf-8'),
        ]

    def _request(self, method, url, get=None, post=None, auth=False):
        content_type = None

        if get:
            url += "?%s" % urllib.parse.urlencode(get)
        if post:
            post = urllib.parse.urlencode(post).encode('utf-8')
            content_type = 'application/x-www-form-urlencoded'
            self.msg.debug("POST data: " + str(post))

        self.msg.debug(method + " URL: " + url)
        request = urllib.request.Request(url, post)
        request.get_method = lambda: method

        if content_type:
            request.add_header('Content-Type', content_type)

        if auth:
            request.add_header('Authorization', '{0} {1}'.format(
                self._get_userconfig('token_type').capitalize(),
                self._get_userconfig('access_token'),
            ))

        try:
            response = self.opener.open(request)

            if response.info().get('content-encoding') == 'gzip':
                response = gzip.GzipFile(fileobj=response).read().decode('utf-8')
            else:
                response = response.read().decode('utf-8')

            return json.loads(response)
        except urllib.error.HTTPError as e:
            raise utils.APIError("Connection error: %s" % e)
        except urllib.error.URLError as e:
            raise utils.APIError("URL error: %s" % e)
        except socket.timeout:
            raise utils.APIError("Operation timed out.")

    def _request_access_token(self, refresh=False):
        """
        Requests or refreshes the access token through OAuth2
        """
        params = {
            'client_id':     self.client_id,
        }

        if refresh:
            self.msg.info('Refreshing access token...')

            params['grant_type'] = 'refresh_token'
            params['refresh_token'] = self._get_userconfig('refresh_token')
        else:
            self.msg.info('Requesting access token...')

            params['code'] = self.pin
            params['code_verifier'] = self.code_verifier
            params['grant_type'] = 'authorization_code'

        data = self._request('POST', self.auth_url, post=params)

        timestamp = int(time.time())

        self._set_userconfig('access_token',  data['access_token'])
        self._set_userconfig('token_type',    data['token_type'])
        self._set_userconfig('expires',       timestamp + data['expires_in'])
        self._set_userconfig('refresh_token', data['refresh_token'])

        self.logged_in = True
        self._emit_signal('userconfig_changed')

    def check_credentials(self):
        timestamp = int(time.time())

        if not self._get_userconfig('access_token'):
            self._request_access_token(False)
        elif (timestamp+60) > self._get_userconfig('expires'):
            try:
                self._request_access_token(True)
            except utils.APIError:
                self._request_access_token(False)
        else:
            self.logged_in = True

        if not self.userid:
            self._refresh_user_info()

        return True

    def _refresh_user_info(self):
        self.msg.info('Refreshing user details...')

        data = self._request("GET", self.query_url + "/users/@me", auth=True)

        self._set_userconfig('userid', data['id'])
        self._set_userconfig('username', data['name'])

        self.userid = data['id']
        self._emit_signal('userconfig_changed')

    def fetch_list(self):
        self.check_credentials()
        shows = {}

        fields = 'id,alternative_titles,title,start_date,main_picture,status,' + self.total_str
        listfields = 'score,status,start_date,finish_date,updated_at,' + self.watched_str
        params = {
            'fields': '%s,list_status{%s}' % (fields, listfields),
            'limit': self.library_page_limit,
            'nsfw': 'true'
        }

        url = "{}/users/@me/{}list?{}".format(self.query_url, self.mediatype, urllib.parse.urlencode(params))
        i = 1

        while url:
            self.msg.info('Downloading list (page %d)...' % i)
            data = self._request('GET', url, auth=True)
            for item in data['data']:
                showid = item['node']['id']
                shows[showid] = utils.show()
                shows[showid].update({
                    'id': showid,
                    'title': item['node']['title'],
                    'url': "https://myanimelist.net/%s/%d" % (self.mediatype, showid),
                    'aliases': self._get_aliases(item['node']),
                    'image': item['node'].get('main_picture', {}).get('large'),
                    'image_thumb': item['node'].get('main_picture', {}).get('medium'),
                    'total': item['node'][self.total_str],
                    'status': self._translate_status(item['node']['status']),
                    'start_date': self._str2date(item['node'].get('start_date')),
                    'my_progress': item['list_status'][self.watched_str],
                    'my_score': item['list_status']['score'],
                    'my_status': item['list_status']['status'],
                    'my_start_date': self._str2date(item['list_status'].get('start_date')),
                    'my_finish_date': self._str2date(item['list_status'].get('finish_date')),
                    'my_last_update': self._iso2datetime(item['list_status'].get('updated_at')),
                })

            url = data['paging'].get('next')
            i += 1

        return shows

    def add_show(self, item):
        self.check_credentials()
        self.msg.info("Adding item %s..." % item['title'])
        self._update_entry(item)

    def update_show(self, item):
        self.check_credentials()
        self.msg.info("Updating item %s..." % item['title'])
        return self._iso2datetime(self._update_entry(item)['updated_at'])

    def delete_show(self, item):
        self.check_credentials()
        self.msg.info("Deleting item %s..." % item['title'])
        data = self._request('DELETE', self.query_url + '/%s/%d/my_list_status' % (self.mediatype, item['id']), auth=True)

    def search(self, criteria, method):
        self.check_credentials()
        self.msg.info("Searching for {}...".format(criteria))

        fields = 'alternative_titles,end_date,genres,id,main_picture,mean,media_type,' + self.total_str + ',popularity,rating,start_date,status,studios,synopsis,title'
        params = {'fields': fields, 'nsfw': 'true'}

        if method == utils.SearchMethod.KW:
            url = '/%s' % self.mediatype
            params['q'] = criteria
            params['limit'] = self.search_page_limit
        elif method == utils.SearchMethod.SEASON:
            season, season_year = criteria

            url = '/%s/season/%d/%s' % (self.mediatype, season_year, self.season_translate[season])
            params['limit'] = self.season_page_limit
        else:
            raise utils.APIError("Invalid search method.")

        results = []
        data = self._request('GET', self.query_url + url, get=params, auth=True)
        for item in data['data']:
            results.append(self._parse_info(item['node']))

        self._emit_signal('show_info_changed', results)
        return results

    def request_info(self, itemlist):
        self.check_credentials()
        infolist = []

        fields = 'alternative_titles,end_date,genres,id,main_picture,mean,media_type,' + self.total_str + ',popularity,rating,start_date,status,studios,synopsis,title'
        params = {'fields': fields, 'nsfw': 'true'}
        for item in itemlist:
            data = self._request('GET', self.query_url + '/%s/%d' % (self.mediatype, item['id']), get=params, auth=True)
            infolist.append(self._parse_info(data))

        self._emit_signal('show_info_changed', infolist)
        return infolist

    def _update_entry(self, item):
        values = {}
        if 'my_progress' in item:
            values[self.watched_send_str] = item['my_progress']
        if 'my_status' in item:
            values['status'] = item['my_status']
        if 'my_score' in item:
            values['score'] = item['my_score']
        if 'my_start_date' in item:
            values['start_date'] = item['my_start_date'] or ""
        if 'my_finish_date' in item:
            values['finish_date'] = item['my_finish_date'] or ""

        data = self._request('PATCH', self.query_url + '/%s/%d/my_list_status' % (self.mediatype, item['id']), post=values, auth=True)

        return data

    def _get_aliases(self, item):
        aliases = [item['alternative_titles']['en'], item['alternative_titles']['ja']] + item['alternative_titles']['synonyms']

        return aliases

    def _parse_info(self, item):
        info = utils.show()
        showid = item['id']

        info.update({
            'id': showid,
            'title': item['title'],
            'url': "https://myanimelist.net/%s/%d" % (self.mediatype, showid),
            'aliases': self._get_aliases(item),
            'type': self.type_translate.get(item['media_type'], utils.Type.UNKNOWN),
            'total': item[self.total_str],
            'status': self._translate_status(item['status']),
            'image': item.get('main_picture', {}).get('large'),
            'start_date': self._str2date(item.get('start_date')),
            'end_date': self._str2date(item.get('end_date')),
            'extra': [
                ('English',         item['alternative_titles'].get('en')),
                ('Japanese',        item['alternative_titles'].get('ja')),
                ('Synonyms',        item['alternative_titles'].get('synonyms')),
                ('Synopsis',        item.get('synopsis')),
                ('Type',            item.get('media_type')),
                ('Mean score',   item.get('mean')),
                ('Status',          self._translate_status(item['status'])),
            ]
        })

        return info

    def _translate_status(self, orig_status):
        return self.status_translate.get(orig_status, utils.Status.UNKNOWN)

    def _str2date(self, string):
        if string is None:
            return None

        try:
            return datetime.datetime.strptime(string, "%Y-%m-%d")
        except Exception:
            self.msg.debug('Invalid date {}'.format(string))
            return None  # Ignore date if it's invalid

    def _iso2datetime(self, string):
        if string is None:
            return None

        try:
            return datetime.datetime.fromisoformat(string).replace(tzinfo=datetime.timezone.utc)
        except Exception:
            self.msg.debug('Invalid date {}'.format(string))
            return None  # Ignore date if it's invalid



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from datetime import date

from PyQt6 import QtCore
from PyQt6.QtWidgets import (QComboBox, QDialog, QDialogButtonBox, QHBoxLayout, QLineEdit, QMessageBox, QPushButton,
                             QRadioButton, QSpinBox, QSplitter, QStackedWidget, QVBoxLayout)

from trackma import utils
from trackma.ui.qt.details import DetailsDialog
from trackma.ui.qt.widgets import AddCardView, AddTableDetailsView


class AddDialog(QDialog):
    worker = None
    selected_show = None
    results = []

    def __init__(self, parent, worker, current_status, default=None):
        QDialog.__init__(self, parent)
        self.resize(950, 700)
        self.setWindowTitle('Search/Add from Remote')
        self.worker = worker
        self.current_status = current_status
        self.default = default
        if default:
            self.setWindowTitle(
                'Search/Add from Remote for new show: %s' % default)

        # Get available search methods and default to keyword search if not reported by the API
        search_methods = self.worker.engine.mediainfo.get(
            'search_methods', [utils.SearchMethod.KW])

        layout = QVBoxLayout()

        # Create top layout
        top_layout = QHBoxLayout()

        if utils.SearchMethod.KW in search_methods:
            self.search_rad = QRadioButton('By keyword:')
            self.search_rad.setChecked(True)
            self.search_txt = QLineEdit()
            self.search_txt.setClearButtonEnabled(True)
            self.search_txt.returnPressed.connect(self.s_search)
            if default:
                self.search_txt.setText(default)
            self.search_btn = QPushButton('Search')
            self.search_btn.clicked.connect(self.s_search)
            top_layout.addWidget(self.search_rad)
            top_layout.addWidget(self.search_txt)
        else:
            top_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)

        top_layout.addWidget(self.search_btn)

        # Create filter line
        filters_layout = QHBoxLayout()

        if utils.SearchMethod.SEASON in search_methods:
            self.season_rad = QRadioButton('By season:')
            self.season_combo = QComboBox()
            self.season_combo.addItem('Winter', utils.Season.WINTER)
            self.season_combo.addItem('Spring', utils.Season.SPRING)
            self.season_combo.addItem('Summer', utils.Season.SUMMER)
            self.season_combo.addItem('Fall', utils.Season.FALL)

            self.season_year = QSpinBox()

            today = date.today()
            current_season = (today.month - 1) // 3

            self.season_year.setRange(1900, today.year)
            self.season_year.setValue(today.year)
            self.season_combo.setCurrentIndex(current_season)

            filters_layout.addWidget(self.season_rad)
            filters_layout.addWidget(self.season_combo)
            filters_layout.addWidget(self.season_year)

            filters_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeft)
            filters_layout.addWidget(QSplitter())
        else:
            filters_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)

        view_combo = QComboBox()
        view_combo.addItem('Card view')
        view_combo.addItem('Table view')
        view_combo.currentIndexChanged.connect(self.s_change_view)

        filters_layout.addWidget(view_combo)

        # Create central content
        self.contents = QStackedWidget()

        # Set up views
        tableview = AddTableDetailsView(None, self.worker)
        tableview.changed.connect(self.s_selected)

        cardview = AddCardView(api_info=self.worker.engine.api_info)
        cardview.changed.connect(self.s_selected)
        cardview.doubleClicked.connect(self.s_show_details)

        self.contents.addWidget(cardview)
        self.contents.addWidget(tableview)

        # Use for testing
        # self.set_results([{'id': 1, 'title': 'Hola', 'image': 'https://omaera.org/icon.png'}])

        bottom_buttons = QDialogButtonBox()
        bottom_buttons.addButton("Cancel", QDialogButtonBox.ButtonRole.RejectRole)
        self.add_btn = bottom_buttons.addButton(
            "Add", QDialogButtonBox.ButtonRole.AcceptRole)
        self.add_btn.setEnabled(False)
        bottom_buttons.accepted.connect(self.s_add)
        bottom_buttons.rejected.connect(self.close)

        # Finish layout
        layout.addLayout(top_layout)
        layout.addLayout(filters_layout)
        layout.addWidget(self.contents)
        layout.addWidget(bottom_buttons)
        self.setLayout(layout)

        if utils.SearchMethod.SEASON in search_methods:
            self.search_txt.setFocus()

    def worker_call(self, function, ret_function, *args, **kwargs):
        # Run worker in a thread
        self.worker.set_function(function, ret_function, *args, **kwargs)
        self.worker.start()

    def _enable_widgets(self, enable):
        self.search_btn.setEnabled(enable)
        self.contents.currentWidget().setEnabled(enable)

    def set_results(self, results):
        self.results = results
        self.contents.currentWidget().setResults(self.results)

    # Slots
    def s_show_details(self):
        detailswindow = DetailsDialog(self, self.worker, self.selected_show)
        detailswindow.setModal(True)
        detailswindow.show()

    def s_change_view(self, item):
        self.contents.currentWidget().getModel().setResults(None)
        self.contents.setCurrentIndex(item)
        self.contents.currentWidget().getModel().setResults(self.results)

    def s_search(self):
        if self.search_rad.isChecked():
            criteria = self.search_txt.text().strip()
            if not criteria:
                return
            method = utils.SearchMethod.KW
        elif self.season_rad.isChecked():
            criteria = (self.season_combo.itemData(
                self.season_combo.currentIndex()), self.season_year.value())
            method = utils.SearchMethod.SEASON

        self.contents.currentWidget().clearSelection()
        self.selected_show = None

        self._enable_widgets(False)
        self.add_btn.setEnabled(False)

        self.worker_call('search', self.r_searched, criteria, method)

    def s_selected(self, show):
        self.selected_show = show
        self.add_btn.setEnabled(True)

    def s_add(self):
        if self.selected_show:
            self.worker_call('add_show', self.r_added,
                             self.selected_show, self.current_status)

    # Worker responses
    def r_searched(self, result):
        self._enable_widgets(True)

        if result['success']:
            self.set_results(result['result'])

            """
            if self.table.currentRow() == 0:  # Row number hasn't changed but the data probably has!
                self.s_show_selected(self.table.item(0, 0))
            self.table.setCurrentItem(self.table.item(0, 0))"""
        else:
            self.set_results(None)

    def r_added(self, result):
        if result['success']:
            if self.default:
                self.accept()
            else:
                QMessageBox.information(
                    self, 'Information', "Show added successfully")



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import argparse
import cmd as command
import os
import re
import shlex
import subprocess
import sys
import textwrap
from operator import itemgetter  # Used for sorting list

from trackma import messenger
from trackma import utils
from trackma.accounts import AccountManager
from trackma.engine import Engine

try:
    import readline
    has_readline = True
except ImportError:
    has_readline = False

_COLOR_RESET = '\033[0m'
_COLOR_ENGINE = '\033[0;32m'
_COLOR_DATA = '\033[0;33m'
_COLOR_API = '\033[0;34m'
_COLOR_TRACKER = '\033[0;35m'
_COLOR_ERROR = '\033[0;31m'
_COLOR_FATAL = '\033[1;31m'

_COLOR_AIRING = '\033[0;34m'
_COLOR_BEHIND = '\033[0;31m'

# We must use mark these with special characters for
# readline to calculate line width correctly
_PCOLOR_RESET = '\001\033[0m\002'
_PCOLOR_USER = '\001\033[0;32m\002'
_PCOLOR_API = '\001\033[0;34m\002'
_PCOLOR_MEDIATYPE = '\001\033[0;33m\002'
_PCOLOR_FILTER = '\001\033[0;35m\002'


class Trackma_cmd(command.Cmd):
    """
    Main program, inherits from the useful Cmd class
    for interactive console
    """
    engine = None
    filter_num = 1
    sort = 'title'
    completekey = 'Tab'
    cmdqueue = []
    stdout = sys.stdout
    in_prompt = False
    sortedlist = []
    needed_args = {
        'altname':      (1, 2),
        'filter':       (0, 1),
        'sort':         1,
        'mediatype':    (0, 1),
        'info':         1,
        'search':       1,
        'add':          1,
        'del':          1,
        'delete':       1,
        'play':         (1, 2),
        'openfolder':   1,
        'rescan':       (0, 1),
        'update':       (1, 2),
        'score':        2,
        'status':       2,
    }

    def __init__(self, account_num=None, debug=False, interactive=True):
        super().__init__()

        if interactive:
            print('Trackma v'+utils.VERSION+'  Copyright (C) 2012-2026  z411')
            print(
                'This program comes with ABSOLUTELY NO WARRANTY; for details type `about\'')
            print('This is free software, and you are welcome to redistribute it')
            print('under certain conditions; see the COPYING file for details.')
            print()

        self.interactive = interactive
        self.debug = debug

        self.accountman = Trackma_accounts()
        if account_num:
            try:
                self.account = self.accountman.get_account(account_num)
            except KeyError:
                print("Account {} doesn't exist.".format(account_num))
                self.account = self.accountman.select_account(True)
            except ValueError:
                print("Account {} must be numeric.".format(account_num))
                self.account = self.accountman.select_account(True)
        else:
            self.account = self.accountman.select_account(False)

    def forget_account(self):
        self.accountman.set_default(None)

    def _update_prompt(self):
        self.prompt = "{c_u}{u}{c_r} [{c_a}{a}{c_r}] ({c_mt}{mt}{c_r}) {c_s}{s}{c_r} >> ".format(
            u=self.engine.get_userconfig('username'),
            a=self.engine.api_info['shortname'],
            mt=self.engine.api_info['mediatype'],
            s=self.engine.mediainfo['statuses_dict'][self.filter_num].lower().replace(
                ' ', ''),
            c_r=_PCOLOR_RESET,
            c_u=_PCOLOR_USER,
            c_a=_PCOLOR_API,
            c_mt=_PCOLOR_MEDIATYPE,
            c_s=_COLOR_RESET
        )

    def _load_list(self, *args):
        showlist = self.engine.filter_list(self.filter_num)
        sortedlist = sorted(showlist, key=itemgetter(self.sort))
        self.sortedlist = list(enumerate(sortedlist, 1))

    def _get_show(self, title):
        # Attempt parsing list index
        # otherwise use title
        try:
            index = int(title)-1
            return self.sortedlist[index][1]
        except (ValueError, AttributeError, IndexError):
            return self.engine.get_show_info(title=title)

    def _ask_update(self, show, episode):
        do = input(
            "Should I update {} to episode {}? [y/N] ".format(show['title'], episode))
        if do.lower() == 'y':
            self.engine.set_episode(show['id'], episode)

    def _ask_add(self, show, episode):
        do = input(
            "Should I search for the show {}? [y/N] ".format(show['title']))
        if do.lower() == 'y':
            self.do_add([show['title']])

    def start(self):
        """
        Initializes the engine

        Creates an Engine object and starts it.
        """

        if self.interactive:
            print('Initializing engine...')
        self.engine = Engine(self.account, self.messagehandler)
        if not self.interactive:
            self.engine.set_config("tracker_enabled", False)
            self.engine.set_config("library_autoscan", False)
            self.engine.set_config("use_hooks", False)

        self.engine.connect_signal('show_added', self._load_list)
        self.engine.connect_signal('show_deleted', self._load_list)
        self.engine.connect_signal('status_changed', self._load_list)
        self.engine.connect_signal('episode_changed', self._load_list)
        self.engine.connect_signal('prompt_for_update', self._ask_update)
        self.engine.connect_signal('prompt_for_add', self._ask_add)
        self.engine.start()

        # Start with default filter selected
        self.filter_num = self.engine.mediainfo['statuses'][0]
        self._load_list()

        if self.interactive:
            self._update_prompt()

            print()
            print("Ready. Type 'help' for a list of commands.")
            print("Press tab for autocompletion and up/down for command history.")
            self.do_filter(None)  # Show available filters
            print()
        else:
            # We set the message handler only after initializing
            # so we still receive the important messages but avoid
            # the initial spam.
            self.engine.set_message_handler(self.messagehandler)

    def do_about(self, args):
        print("Trackma {}  by z411 (z411@omaera.org)".format(utils.VERSION))
        print("Trackma is an open source client for media tracking websites.")
        print("https://github.com/z411/trackma")
        print()
        print("This program is licensed under the GPLv3 and it comes with ABSOLUTELY NO WARRANTY.")
        print("Many contributors have helped to run this project; for more information see the AUTHORS file.")
        print("For more information about the license, see the COPYING file.")
        print()
        print("If you encounter any problems please report them in https://github.com/z411/trackma/issues")
        print()
        print("This is the CLI version of Trackma. To see available commands type `help'.")
        print("For other available interfaces please see the README file.")
        print()

    def do_help(self, arg):
        if arg:
            try:
                doc = getattr(self, 'do_' + arg).__doc__
                if doc:
                    (name, args, expl, usage, examples) = self._parse_doc(arg, doc)

                    print()
                    print(name)
                    for line in expl:
                        print("  {}".format(line))
                    if args:
                        print("\n  Arguments:")
                        for arg in args:
                            if arg[2]:
                                print("    {}: {}".format(arg[0], arg[1]))
                            else:
                                print("    {} (optional): {}".format(
                                    arg[0], arg[1]))
                    if usage:
                        print("\n  Usage: " + usage)
                    for example in examples:
                        print("  Example: " + example)
                    print()
                    return
            except AttributeError:
                pass

            print("No help available.")
            return
        else:
            CMD_LENGTH = 11
            ARG_LENGTH = 13

            (height, width) = utils.get_terminal_size()
            prev_width = CMD_LENGTH + ARG_LENGTH + 3

            tw = textwrap.TextWrapper()
            tw.width = width - 2
            tw.subsequent_indent = ' ' * prev_width

            print()
            print(" {0:>{1}} {2:{3}} {4}".format(
                'command', CMD_LENGTH,
                'args', ARG_LENGTH,
                'description'))
            print(" " + "-"*(min(prev_width+81, width-3)))

            names = self.get_names()
            names.sort()
            for name in names:
                if name[:3] == 'do_':
                    doc = getattr(self, name).__doc__
                    if not doc:
                        continue

                    cmd = name[3:]
                    (name, args, expl, usage, examples) = self._parse_doc(cmd, doc)

                    line = " {0:>{1}} {2:{3}} {4}".format(
                        name, CMD_LENGTH,
                        '<' + ','.join(a[0] for a in args) + '>', ARG_LENGTH,
                        expl[0])
                    print(tw.fill(line))

            print()
            print("Use `help <command>` for detailed information.")
            print()

    def do_account(self, args):
        """
        Switch to a different account.
        """

        self.account = self.accountman.select_account(True)
        self.engine.reload(account=self.account)

        # Start with default filter selected
        self.filter_num = self.engine.mediainfo['statuses'][0]
        self._load_list()
        self._update_prompt()

    def do_filter(self, args):
        """
        Changes the filtering of list by status (shows current if empty).

        :optparam status Name of status to filter
        :usage filter [filter type]
        """
        # Query the engine for the available statuses
        # that the user can choose
        if args:
            try:
                self.filter_num = self._guess_status(args[0].lower())
                self._load_list()
                self._update_prompt()
            except KeyError:
                print("Invalid filter.")
        else:
            print("Available statuses: %s" % ', '.join(v.lower().replace(' ', '')
                                                       for v in self.engine.mediainfo['statuses_dict'].values()))

    def do_sort(self, args):
        """
        Change of the lists

        :param type Sort type; available types: id, title, my_progress, total, my_score
        :usage sort <sort type>
        """
        sorts = ('id', 'title', 'my_progress', 'total', 'my_score')
        if args[0] in sorts:
            self.sort = args[0]
            self._load_list()
        else:
            print("Invalid sort.")

    def do_mediatype(self, args):
        """
        Reloads engine with different mediatype (shows current if empty).
        Call with no arguments to see supported mediatypes.

        :optparam mediatype Mediatype name
        :usage mediatype [mediatype]
        """
        if args:
            if args[0] in self.engine.api_info['supported_mediatypes']:
                self.engine.reload(mediatype=args[0])

                # Start with default filter selected
                self.filter_num = self.engine.mediainfo['statuses'][0]
                self._load_list()
                self._update_prompt()
            else:
                print("Invalid mediatype.")
        else:
            print("Supported mediatypes: %s" % ', '.join(
                self.engine.api_info['supported_mediatypes']))

    def do_ls(self, args):
        self.do_list(args)

    def do_list(self, args):
        """
        Lists all shows available in the local list.

        :name list|ls
        """
        # Show the list in memory
        self._make_list(self.sortedlist)

    def do_info(self, args):
        """
        Gets detailed information about a local show.

        :param show Show index or title.
        :usage info <show index or title>
        """
        try:
            show = self._get_show(args[0])
            details = self.engine.get_show_details(show)
        except utils.TrackmaError as e:
            self.display_error(e)
            return

        altnames = self.engine.altnames()
        altname = altnames.get(show['id'])
        print(show['title'])
        print("-" * len(show['title']))
        print(show['url'])
        print()
        if altname := altnames.get(show['id']):
            print(f"Altname: {altname}")

        for line in details['extra']:
            print("%s: %s" % line)

    def do_search(self, args):
        """
        Does a regex search on shows in the local lists, including synonyms and altname.

        :param pattern Regex pattern to search for.
        :usage search <pattern>
        """
        compiled_pattern = re.compile(args[0], re.I)
        altnames = self.engine.altnames()

        def matches_show(show):
            titles = [show['title'], *show['aliases']]
            if altname := altnames.get(show['id']):
                titles.append(altname)
            return any(map(compiled_pattern.search, titles))

        sublist = [v for v in self.sortedlist if matches_show(v[1])]
        self._make_list(sublist)

    def do_add(self, args):
        """
        Search for a show in the remote service and add it.

        :param pattern Show criteria to search.
        :usage add <pattern>
        """
        try:
            entries = self.engine.search(args[0])
        except utils.TrackmaError as e:
            self.display_error(e)
            return

        for i, entry in enumerate(entries, start=1):
            print("%d: (%s) %s" % (i, entry['type'], entry['title']))
        do_update = input("Choose show to add (blank to cancel): ")
        if do_update != '':
            try:
                show = entries[int(do_update)-1]
            except ValueError:
                print("Choice must be numeric.")
                return
            except IndexError:
                print("Invalid show.")
                return

            # Tell the engine to add the show
            try:
                self.engine.add_show(show, self.filter_num)
            except utils.TrackmaError as e:
                self.display_error(e)

    def do_del(self, args):
        self.do_delete(args)

    def do_delete(self, args):
        """
        Deletes a show from the local list.

        :name delete|del
        :param show Show index or title.
        :usage delete <show index or title>
        """
        try:
            show = self._get_show(args[0])

            do_delete = input("Delete %s? [y/N] " % show['title'])
            if do_delete.lower() == 'y':
                self.engine.delete_show(show)
        except utils.TrackmaError as e:
            self.display_error(e)

    def do_rescan(self, args):
        """
        Re-scans the local library.

        :optparam path Base path for the scan. Defaults to all library folders if omitted.
        :usage rescan [path]
        """
        path = args[0] if args else None
        self.engine.scan_library(rescan=True, path=path)

    def do_random(self, args):
        """
        Starts the media player with a random new episode.
        """
        try:
            args = self.engine.play_random()
            utils.spawn_process(args)
        except utils.TrackmaError as e:
            self.display_error(e)

    def do_tracker(self, args):
        """
        Shows information about the tracker, if it's running.

        :usage tracker
        """
        try:
            info = self.engine.tracker_status()
            print("- Tracker status -")

            if info:
                if info['state'] == utils.Tracker.NOVIDEO:
                    state = 'No video'
                elif info['state'] == utils.Tracker.PLAYING:
                    state = 'Playing'
                elif info['state'] == utils.Tracker.UNRECOGNIZED:
                    state = 'Unrecognized'
                elif info['state'] == utils.Tracker.NOT_FOUND:
                    state = 'Not found'
                elif info['state'] == utils.Tracker.IGNORED:
                    state = 'Ignored'
                else:
                    state = 'N/A'

                print("State: {}".format(state))
                print("Filename: {}".format(info['filename'] or 'N/A'))
                print("Timer: {}{}".format(
                    info['timer'] or 'N/A', ' [P]' if info['paused'] else ''))
                if info['show']:
                    (show, ep) = info['show']
                    print("Show: {}\nEpisode: {}".format(show['title'], ep))
                else:
                    print("Show: N/A")
            else:
                print("Not started")
        except utils.TrackmaError as e:
            self.display_error(e)

    def do_play(self, args):
        """
        Starts the media player with the specified episode number (next if unspecified).

        :param show Episode index or title.
        :optparam ep Episode number. Assume next if not specified.
        :usage play <show index or title> [episode number]
        """
        try:
            episode = 0
            show = self._get_show(args[0])

            # If the user specified an episode, play it
            # otherwise play the next episode not watched yet
            if len(args) > 1:
                episode = args[1]

            args = self.engine.play_episode(show, episode)
            utils.spawn_process(args)
        except utils.TrackmaError as e:
            self.display_error(e)

    def do_openfolder(self, args):
        """
        Opens the folder containing the show

        :param show Show index or name.
        :usage openfolder <show index or name>
        """

        try:
            show = self._get_show(args[0])
            filename = self.engine.get_episode_path(show)
            with open(os.devnull, 'wb') as DEVNULL:
                if sys.platform == 'darwin':
                    subprocess.Popen(["open",
                                      os.path.dirname(filename)], stdout=DEVNULL, stderr=DEVNULL)
                elif sys.platform == 'win32':
                    subprocess.Popen(["explorer",
                                      os.path.dirname(filename)], stdout=DEVNULL, stderr=DEVNULL)
                else:
                    subprocess.Popen(["xdg-open",
                                      os.path.dirname(filename)], stdout=DEVNULL, stderr=DEVNULL)
        except OSError:
            # xdg-open failed.
            self.display_error("Could not open folder.")
        except utils.TrackmaError as e:
            self.display_error(e)

    def do_update(self, args):
        """
        Updates the progress of a show to the specified episode (next if unspecified).

        :param show Show index, title or filename (prepend with file:).
        :optparam ep Episode number (numeric).
        :usage update <show index or name> [episode number]
        :example update Toradora! 5
        :example update 6
        :example update file:filename.mkv
        """
        try:
            if args[0][:5] == "file:":
                (show, ep) = self.engine.get_show_info(filename=args[0][5:])
            else:
                (show, ep) = (self._get_show(args[0]), None)

            if len(args) > 1:
                self.engine.set_episode(show['id'], args[1])
            else:
                self.engine.set_episode(
                    show['id'], ep or show['my_progress']+1)
        except IndexError:
            print("Missing arguments.")
        except utils.TrackmaError as e:
            self.display_error(e)

    def do_score(self, args):
        """
        Changes the score of a show.

        :param show Show index or name.
        :param score Score to set (numeric/decimal).
        :usage score <show index or name> <score>
        """
        try:
            show = self._get_show(args[0])
            self.engine.set_score(show['id'], args[1])
        except IndexError:
            print("Missing arguments.")
        except utils.TrackmaError as e:
            self.display_error(e)

    def do_status(self, args):
        """
        Changes the status of a show.
        Use the command `filter` without arguments to see the available statuses.

        :param show Show index or name.
        :param status Status name. Use `filter` without args to list them.
        :usage status <show index or name> <status name>
        """
        try:
            _showtitle = args[0]
            _filter = args[1]
        except IndexError:
            print("Missing arguments.")
            return

        try:
            _filter_num = self._guess_status(_filter)
        except KeyError:
            print("Invalid filter.")
            return

        try:
            show = self._get_show(_showtitle)
            self.engine.set_status(show['id'], _filter_num)
        except utils.TrackmaError as e:
            self.display_error(e)

    def do_altname(self, args):
        """
        Changes the alternative name of a show (removes if unspecified).
        Use the command 'altname' without arguments to clear the alternative
        name.

        :param show Show index or name
        :param alt  The alternative name. Use `altname` without alt to clear it
        :usage altname <show index or name> <alternative name>
        """
        try:
            show = self._get_show(args[0])
            altname = args[1] if len(args) > 1 else ''
            self.engine.altname(show['id'], altname)
        except IndexError:
            print("Missing arguments")
            return
        except utils.TrackmaError as e:
            self.display_error(e)

    def do_send(self, args):
        """
        Sends queued changes to the remote service.
        """
        try:
            self.engine.list_upload()
        except utils.TrackmaError as e:
            self.display_error(e)

    def do_retrieve(self, args):
        """
        Retrieves the remote list overwrites the local one.
        """
        try:
            if self.engine.get_queue():
                answer = input(
                    "There are unqueued changes. Overwrite local list? [y/N] ")
                if answer.lower() == 'y':
                    self.engine.list_download()
            else:
                self.engine.list_download()
            self._load_list()
        except utils.TrackmaError as e:
            self.display_error(e)

    def do_clearqueue(self, args):
        """
        Undo all changes in queue.
        """
        try:
            self.engine.queue_clear()
        except utils.TrackmaError as e:
            self.display_error(e)

    def do_viewqueue(self, args):
        """
        List the queued changes.
        """
        queue = self.engine.get_queue()
        if queue:
            print("Queue:")
            for show in queue:
                print("- %s" % show['title'])
        else:
            print("Queue is empty.")

    def do_exit(self, args):
        self.do_quit(args)

    def do_quit(self, args):
        """
        Quits the program.

        :name quit|exit
        """
        try:
            self.engine.unload()
        except utils.TrackmaError as e:
            self.display_error(e)

        print('Bye!')
        sys.exit(0)

    def do_EOF(self, args):
        print()
        self.do_quit(args)

    def _complete_show(self, text, line, begidx, endidx):
        if text:
            return self.engine.regex_list_titles(text)

    complete_info = _complete_show
    complete_altname = _complete_show
    complete_update = _complete_show
    complete_play = _complete_show
    complete_score = _complete_show
    complete_status = _complete_show
    complete_delete = _complete_show

    def complete_filter(self, text, line, begidx, endidx):
        all_items = [v.lower().replace(' ', '') for v in self.engine.mediainfo['statuses_dict'].values()]
        return [i for i in all_items if i.startswith(text)]

    def parse_args(self, arg):
        if arg:
            return shlex.split(arg)

        return []

    def emptyline(self):
        return

    def preloop(self):
        """ Override. """
        self.in_prompt = True

    def precmd(self, line):
        """ Override. """
        self.in_prompt = False
        return line

    def postcmd(self, stop, line):
        """ Override. """
        self.in_prompt = True
        return stop

    def onecmd(self, line):
        """ Override. """
        cmd, arg, line = self.parseline(line)
        if not line:
            return self.emptyline()
        if cmd is None:
            return self.default(line)
        self.lastcmd = line
        if line == 'EOF':
            self.lastcmd = ''
        if cmd == '':
            return self.default(line)
        elif cmd == 'help':
            return self.do_help(arg)
        else:
            try:
                args = self.parse_args(arg)
            except ValueError:
                return self.default(line)
            else:
                return self.execute(cmd, args, line)

    def execute(self, cmd, args, line):
        try:
            func = getattr(self, 'do_' + cmd)
        except AttributeError:
            return self.default(line)

        try:
            needed = self.needed_args[cmd]
        except KeyError:
            needed = 0

        if isinstance(needed, int):
            needed = (needed, needed)

        if needed[0] <= len(args) <= needed[1]:
            return func(args)
        else:
            print("Incorrect number of arguments. See `help %s`" % cmd)

    def display_error(self, e):
        print("%s%s: %s%s" % (_COLOR_ERROR, type(e).__name__, e, _COLOR_RESET))

    def messagehandler(self, classname, msgtype, msg):
        """
        Handles and shows messages coming from
        the engine messenger to provide feedback.
        """
        color_escape = ''
        color_reset = _COLOR_RESET

        if classname == 'Engine':
            color_escape = _COLOR_ENGINE
        elif classname == 'Data':
            color_escape = _COLOR_DATA
        elif classname.startswith('lib'):
            color_escape = _COLOR_API
        elif classname.startswith('Tracker'):
            color_escape = _COLOR_TRACKER
        else:
            color_reset = ''

        if msgtype == messenger.TYPE_INFO:
            out = "%s%s: %s%s" % (color_escape, classname, msg, color_reset)
        elif msgtype == messenger.TYPE_WARN:
            out = "%s%s warning: %s%s" % (
                color_escape, classname, msg, color_reset)
        elif self.debug and msgtype == messenger.TYPE_DEBUG:
            out = "[D] %s%s: %s%s" % (
                color_escape, classname, msg, color_reset)
        else:
            return  # Unrecognized message, don't show anything

        if has_readline and self.in_prompt:
            # If we're in a prompt and receive a message
            # (often from the tracker) we need to clear the line
            # first, show the message, then re-show the prompt.
            buf = readline.get_line_buffer()
            self.stdout.write('\r' + ' '*(len(self.prompt)+len(buf)) + '\r')

            print(out)

            self.stdout.write(self.prompt + buf)
            self.stdout.flush()
        else:
            print(out)

    def _guess_status(self, string):
        for k, v in self.engine.mediainfo['statuses_dict'].items():
            if string.lower() == v.lower().replace(' ', ''):
                return k
        raise KeyError

    def _parse_doc(self, cmd, doc):
        lines = doc.split('\n')
        name = cmd
        args = []
        expl = []
        usage = None
        examples = []

        for line in lines:
            line = line.strip()
            if line[:6] == ":param":
                args.append(line[7:].split(' ', 1) + [True])
            elif line[:9] == ":optparam":
                args.append(line[10:].split(' ', 1) + [False])
            elif line[:6] == ':usage':
                usage = line[7:]
            elif line[:5] == ':name':
                name = line[6:]
            elif line[:8] == ':example':
                examples.append(line[9:])
            elif line:
                expl.append(line)

        return (name, args, expl, usage, examples)

    def _make_list(self, showlist):
        """
        Helper function for printing a formatted show list
        """
        # Fixed column widths
        col_id_length = 7
        col_index_length = 6
        col_title_length = 5
        col_episodes_length = 9
        col_score_length = 6
        altnames = self.engine.altnames()

        # Calculate maximum width for the title column
        # based on the width of the terminal
        (height, width) = utils.get_terminal_size()
        max_title_length = width - col_id_length - \
            col_episodes_length - col_score_length - col_index_length - 5

        # Find the widest title so we can adjust the title column
        for index, show in showlist:
            if len(show['title']) > col_title_length:
                if len(show['title']) > max_title_length:
                    # Stop if we exceeded the maximum column width
                    col_title_length = max_title_length
                    break
                else:
                    col_title_length = len(show['title'])

        # Print header
        print("| {0:{1}} {2:{3}} {4:{5}} {6:{7}} |".format(
            'Index',    col_index_length,
            'Title',    max_title_length,
            'Progress', col_episodes_length,
            'Score',    col_score_length))

        # List shows
        for index, show in showlist:
            if self.engine.mediainfo['has_progress']:
                episodes_str = "{0:3} / {1}".format(
                    show['my_progress'], show['total'] or '?')
            else:
                episodes_str = "-"

            # Get title (and alt. title) and if need be, truncate it
            title_str = show['title']
            if altnames.get(show['id']):
                title_str += " [{}]".format(altnames.get(show['id']))
            title_str = title_str[:max_title_length] if len(
                title_str) > max_title_length else title_str

            # Color title according to status
            if show['status'] == utils.Status.AIRING:
                estimate = utils.estimate_aired_episodes(show)
                if estimate and show['my_progress'] < estimate:
                    # User is behind the (estimated) aired episode
                    colored_title = _COLOR_BEHIND + title_str + _COLOR_RESET
                else:
                    colored_title = _COLOR_AIRING + title_str + _COLOR_RESET
            else:
                colored_title = title_str

            print("| {0:^{1}} {2}{3} {4:{5}} {6:^{7}} |".format(
                index, col_index_length,
                colored_title,
                '.' * (max_title_length-len(title_str)),
                episodes_str, col_episodes_length,
                show['my_score'], col_score_length))

        # Print result count
        print('%d results' % len(showlist))
        print()


class Trackma_accounts(AccountManager):
    def _get_id(self, index):
        if index < 1:
            raise IndexError

        return index

    def select_account(self, bypass):
        if not bypass and self.get_default():
            return self.get_default()
        if self.get_default():
            self.set_default(None)

        while True:
            print('--- Accounts ---')
            self.list_accounts()
            key = input(
                "Input account number ([r#]emember, [a]dd, [c]ancel, [d]elete, [q]uit): ")

            if key.lower() == 'a':
                available_libs = ', '.join(sorted(utils.available_libs.keys()))

                print("--- Add account ---")
                import getpass
                api = input('Enter API (%s): ' % available_libs)
                extra = {}
                try:
                    selected_api = utils.available_libs[api]
                except KeyError:
                    print("Invalid API.")
                    continue

                if selected_api[2] == utils.Login.PASSWD:
                    username = input('Enter username: ')
                    password = getpass.getpass('Enter password (no echo): ')
                elif selected_api[2] in [utils.Login.OAUTH, utils.Login.OAUTH_PKCE]:
                    username = input('Enter account name: ')

                    auth_url = selected_api[3]
                    if selected_api[2] == utils.Login.OAUTH_PKCE:
                        extra['code_verifier'] = utils.oauth_generate_pkce()
                        auth_url = auth_url % extra['code_verifier']

                    print('OAuth Authentication')
                    print('--------------------')
                    print('This website requires OAuth authentication.')
                    print('Please go to the following URL with your browser,')
                    print('follow the steps and paste the given PIN code here.')
                    print()
                    print(auth_url)
                    print()
                    password = input('PIN: ')

                try:
                    self.add_account(username, password, api, extra)
                    print('Done.')
                except utils.AccountError as e:
                    print('Error: %s' % e)
            elif key.lower() == 'd':
                print("--- Delete account ---")
                num = input('Account number to delete: ')
                try:
                    num = int(num)
                    account_id = self._get_id(num)
                    confirm = input("Are you sure you want to delete account %d (%s)? [y/N] " % (
                        num, self.get_account(account_id)['username']))
                    if confirm.lower() == 'y':
                        self.delete_account(account_id)
                        print('Account %d deleted.' % num)
                except ValueError:
                    print("Invalid value.")
                except IndexError:
                    print("Account doesn't exist.")
            elif key.lower() == 'q':
                sys.exit(0)
            else:
                try:
                    if key[0] == 'r':
                        key = key[1:]
                        remember = True
                    else:
                        remember = False

                    num = int(key)
                    account_id = self._get_id(num)
                    if remember:
                        self.set_default(account_id)

                    return self.get_account(account_id)
                except ValueError:
                    print("Invalid value.")
                except (IndexError, KeyError):
                    print("Account doesn't exist.")

    def list_accounts(self):
        accounts = self.get_accounts()

        print("Available accounts:")
        if accounts:
            for k, account in accounts:
                print("%i: %s (%s)" % (k, account['username'], account['api']))
        else:
            print("No accounts.")


def main():
    # Process args
    parser = argparse.ArgumentParser()
    parser.add_argument('-a', '--account', type=int,
                        help='Use specific account number.')
    parser.add_argument('-d', '--debug', action='store_true',
                        help='Show debugging messages.')
    parser.add_argument(
        'cmd', nargs='?', help='Run the following command and exit. Will run in interactive mode if not specified. - will take in commands from stdin.')
    parser.add_argument('args', nargs=argparse.REMAINDER,
                        help='Arguments for the aforementioned command, if any.')
    args = parser.parse_args()

    # Boot Trackma CLI
    main_cmd = Trackma_cmd(args.account, args.debug,
                           interactive=args.cmd is None)
    try:
        main_cmd.start()
        if args.cmd:
            if args.cmd == '-':
                # Run commands from stdin
                for line in sys.stdin:
                    main_cmd.onecmd(line)
            else:
                # Run the specified command in the arguments
                main_cmd.execute(args.cmd, args.args, args.cmd)
        else:
            main_cmd.cmdloop()
    except utils.TrackmaFatal as e:
        main_cmd.forget_account()
        print("%s%s: %s%s" % (_COLOR_FATAL, type(e).__name__, e, _COLOR_RESET))
    except KeyboardInterrupt:
        print()
        print("Interrupted by user.")
        main_cmd.onecmd('EOF')  # TODO prints another prompt before exiting


if __name__ == '__main__':
    main()



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os

from gi.repository import Gdk, Gtk, Pango

from trackma import utils
from trackma.ui.gtk import gtk_dir
from trackma.ui.gtk.statusicon import TrackmaStatusIcon


def reprColor(gdkColor):
    return '#%02x%02x%02x' % (
        round(gdkColor.red_float * 255),
        round(gdkColor.green_float * 255),
        round(gdkColor.blue_float * 255))


def getColor(colorString):
    # Takes a color string in either #RRGGBB format
    # TODO: Take a group, role format (using GTK int values)
    # Returns gdk color
    if colorString[0] == "#":
        return Gdk.color_parse(colorString)

    return Gdk.color_parse("#000000")


@Gtk.Template.from_file(os.path.join(gtk_dir, 'data/settingswindow.ui'))
class SettingsWindow(Gtk.Window):

    __gtype_name__ = 'SettingsWindow'

    btn_save = Gtk.Template.Child()
    switch_tracker = Gtk.Template.Child()

    radio_tracker_local = Gtk.Template.Child()
    radio_tracker_mpris = Gtk.Template.Child()
    entry_player_process = Gtk.Template.Child()
    btn_file_chooser_executable = Gtk.Template.Child()
    listbox_directories = Gtk.Template.Child()
    btn_add_directory = Gtk.Template.Child()
    checkbox_library_startup = Gtk.Template.Child()
    checkbox_library_entire_list = Gtk.Template.Child()
    checkbox_library_full_path = Gtk.Template.Child()

    radio_tracker_plex = Gtk.Template.Child()
    entry_plex_host = Gtk.Template.Child()
    spin_plex_port = Gtk.Template.Child()
    entry_plex_username = Gtk.Template.Child()
    entry_plex_password = Gtk.Template.Child()
    checkbox_plex_obey_wait = Gtk.Template.Child()
    checkbox_plex_ssl = Gtk.Template.Child()
    spin_tracker_update_wait = Gtk.Template.Child()

    radio_tracker_jellyfin = Gtk.Template.Child()
    entry_jellyfin_host = Gtk.Template.Child()
    spin_jellyfin_port = Gtk.Template.Child()
    entry_jellyfin_username = Gtk.Template.Child()
    entry_jellyfin_api_key = Gtk.Template.Child()

    checkbox_tracker_update_close = Gtk.Template.Child()
    checkbox_tracker_update_prompt = Gtk.Template.Child()
    checkbox_tracker_not_found_prompt = Gtk.Template.Child()

    radiobutton_download_days = Gtk.Template.Child()
    radiobutton_download_always = Gtk.Template.Child()
    radiobutton_download_off = Gtk.Template.Child()

    radiobutton_upload_minutes = Gtk.Template.Child()
    radiobutton_upload_size = Gtk.Template.Child()
    radiobutton_upload_always = Gtk.Template.Child()
    radiobutton_upload_off = Gtk.Template.Child()
    checkbox_upload_exit = Gtk.Template.Child()

    spinbutton_download_days = Gtk.Template.Child()
    spinbutton_upload_minutes = Gtk.Template.Child()
    spinbutton_upload_size = Gtk.Template.Child()

    checkbox_auto_status_change = Gtk.Template.Child()
    checkbox_auto_status_change_if_scored = Gtk.Template.Child()
    checkbox_auto_date_change = Gtk.Template.Child()

    checkbox_show_tray = Gtk.Template.Child()
    checkbox_close_to_tray = Gtk.Template.Child()
    checkbox_start_in_tray = Gtk.Template.Child()
    checkbox_tray_api_icon = Gtk.Template.Child()
    checkbox_remember_geometry = Gtk.Template.Child()
    checkbox_classic_progress = Gtk.Template.Child()

    colorbutton_rows_playing = Gtk.Template.Child()
    colorbutton_rows_queued = Gtk.Template.Child()
    colorbutton_rows_new_episode = Gtk.Template.Child()
    colorbutton_rows_is_airing = Gtk.Template.Child()
    colorbutton_rows_not_aired = Gtk.Template.Child()

    colorbutton_progress_bg = Gtk.Template.Child()
    colorbutton_progress_fg = Gtk.Template.Child()
    colorbutton_progress_sub_bg = Gtk.Template.Child()
    colorbutton_progress_sub_fg = Gtk.Template.Child()
    colorbutton_progress_complete = Gtk.Template.Child()

    def __init__(self, engine, config, configfile, transient_for=None):
        Gtk.Window.__init__(self, transient_for=transient_for)
        self.init_template()

        self.engine = engine
        self.config = config
        self.configfile = configfile

        self._color_buttons = {
            'is_playing':        self.colorbutton_rows_playing,
            'is_queued':         self.colorbutton_rows_queued,
            'new_episode':       self.colorbutton_rows_new_episode,
            'is_airing':         self.colorbutton_rows_is_airing,
            'not_aired':         self.colorbutton_rows_not_aired,
            'progress_bg':       self.colorbutton_progress_bg,
            'progress_fg':       self.colorbutton_progress_fg,
            'progress_sub_bg':   self.colorbutton_progress_sub_bg,
            'progress_sub_fg':   self.colorbutton_progress_sub_fg,
            'progress_complete': self.colorbutton_progress_complete
        }

        if os.sys.platform == 'linux':
            self.radio_tracker_mpris.set_visible(True)

        self.radiobutton_download_days.connect(
            "toggled", self._button_toggled, self.spinbutton_download_days)
        self.radiobutton_upload_minutes.connect(
            "toggled", self._button_toggled, self.spinbutton_upload_minutes)
        self.radiobutton_upload_size.connect(
            "toggled", self._button_toggled, self.spinbutton_upload_size)
        self.checkbox_auto_status_change.connect(
            "toggled", self._button_toggled, self.checkbox_auto_status_change_if_scored)
        self.checkbox_show_tray.connect(
            "toggled", self._button_toggled, self.checkbox_close_to_tray)
        self.checkbox_show_tray.connect(
            "toggled", self._button_toggled, self.checkbox_start_in_tray)
        self.checkbox_show_tray.connect(
            "toggled", self._button_toggled, self.checkbox_tray_api_icon)

        self._load_config()

    def _load_config(self):
        """Engine Configuration"""
        self.switch_tracker.set_active(
            self.engine.get_config('tracker_enabled'))

        if self.engine.get_config('tracker_type') == 'local':
            self.radio_tracker_local.set_active(True)
        elif self.engine.get_config('tracker_type') == 'mpris':
            self.radio_tracker_mpris.set_active(True)
        elif self.engine.get_config('tracker_type') == 'plex':
            self.radio_tracker_plex.set_active(True)
        elif self.engine.get_config('tracker_type') == 'jellyfin':
            self.radio_tracker_jellyfin.set_active(True)

        self.entry_player_process.set_text(
            self.engine.get_config('tracker_process'))
        self.btn_file_chooser_executable.set_filename(
            self.engine.get_config('player'))
        self.checkbox_library_startup.set_active(
            self.engine.get_config('library_autoscan'))
        self.checkbox_library_entire_list.set_active(
            self.engine.get_config('scan_whole_list'))
        self.checkbox_library_full_path.set_active(
            self.engine.get_config('library_full_path'))
        self._load_directories(self.engine.get_config('searchdir'))

        self.entry_plex_host.set_text(self.engine.get_config('plex_host'))
        self.spin_plex_port.set_value(int(self.engine.get_config('plex_port')))
        self.checkbox_plex_ssl.set_active(self.engine.get_config('plex_ssl'))
        self.entry_plex_username.set_text(self.engine.get_config('plex_user'))
        self.entry_plex_password.set_text(
            self.engine.get_config('plex_passwd'))
        self.checkbox_plex_obey_wait.set_active(
            self.engine.get_config('plex_obey_update_wait_s'))

        self.entry_jellyfin_host.set_text(
            self.engine.get_config('jellyfin_host'))
        self.spin_jellyfin_port.set_value(
            int(self.engine.get_config('jellyfin_port')))
        self.entry_jellyfin_username.set_text(
            self.engine.get_config('jellyfin_user'))
        self.entry_jellyfin_api_key.set_text(
            self.engine.get_config('jellyfin_api_key'))

        self.spin_tracker_update_wait.set_value(
            self.engine.get_config('tracker_update_wait_s'))
        self.checkbox_tracker_update_close.set_active(
            self.engine.get_config('tracker_update_close'))
        self.checkbox_tracker_update_prompt.set_active(
            self.engine.get_config('tracker_update_prompt'))
        self.checkbox_tracker_not_found_prompt.set_active(
            self.engine.get_config('tracker_not_found_prompt'))

        if self.engine.get_config('autoretrieve') == 'always':
            self.radiobutton_download_always.set_active(True)
        elif self.engine.get_config('autoretrieve') == 'days':
            self.radiobutton_download_days.set_active(True)
        else:
            self.radiobutton_download_off.set_active(True)

        if self.engine.get_config('autosend') == 'always':
            self.radiobutton_upload_always.set_active(True)
        elif self.engine.get_config('autosend') in ('minutes', 'hours'):
            self.radiobutton_upload_minutes.set_active(True)
        elif self.engine.get_config('autosend') == 'size':
            self.radiobutton_upload_size.set_active(True)
        else:
            self.radiobutton_upload_off.set_active(True)

        self.checkbox_upload_exit.set_active(
            self.engine.get_config('autosend_at_exit'))

        self.spinbutton_download_days.set_value(
            self.engine.get_config('autoretrieve_days'))
        self.spinbutton_upload_minutes.set_value(
            self.engine.get_config('autosend_minutes'))
        self.spinbutton_upload_size.set_value(
            self.engine.get_config('autosend_size'))

        self.checkbox_auto_status_change.set_active(
            self.engine.get_config('auto_status_change'))
        self.checkbox_auto_status_change_if_scored.set_active(
            self.engine.get_config('auto_status_change_if_scored'))
        self.checkbox_auto_date_change.set_active(
            self.engine.get_config('auto_date_change'))

        self.checkbox_show_tray.set_active(self.config['show_tray'])
        self.checkbox_close_to_tray.set_active(self.config['close_to_tray'])
        self.checkbox_start_in_tray.set_active(self.config['start_in_tray'])
        self.checkbox_tray_api_icon.set_active(self.config['tray_api_icon'])

        """GTK Interface configuration"""
        self.checkbox_remember_geometry.set_active(
            self.config['remember_geometry'])
        self.checkbox_classic_progress.set_active(
            not self.config['episodebar_style'])

        for color_key, color_button in self._color_buttons.items():
            color = getColor(self.config['colors'][color_key])
            color_button.set_color(color)

        self._set_tracker_radio_buttons()
        self._button_toggled(self.radiobutton_download_days,
                             self.spinbutton_download_days)
        self._button_toggled(self.radiobutton_upload_minutes,
                             self.spinbutton_upload_minutes)
        self._button_toggled(self.radiobutton_upload_size,
                             self.spinbutton_upload_size)
        self._button_toggled(self.checkbox_auto_status_change,
                             self.checkbox_auto_status_change_if_scored)
        self._button_toggled(self.checkbox_show_tray,
                             self.checkbox_close_to_tray)
        self._button_toggled(self.checkbox_show_tray,
                             self.checkbox_start_in_tray)
        self._button_toggled(self.checkbox_show_tray,
                             self.checkbox_tray_api_icon)

        if not TrackmaStatusIcon.is_tray_available():
            self.checkbox_show_tray.set_label(
                'Show tray icon (Not supported in this environment)')
            self.checkbox_show_tray.set_sensitive(False)
            self.checkbox_close_to_tray.set_sensitive(False)
            self.checkbox_start_in_tray.set_sensitive(False)
            self.checkbox_tray_api_icon.set_sensitive(False)

    def _button_toggled(self, widget, spin):
        spin.set_sensitive(widget.get_active())

    @Gtk.Template.Callback()
    def _on_btn_save_clicked(self, btn):
        self.save_config()
        self.destroy()

    @Gtk.Template.Callback()
    def _on_switch_tracker_state_set(self, switch, state):
        self.radio_tracker_local.set_sensitive(state)
        self.radio_tracker_mpris.set_sensitive(state)
        self.radio_tracker_plex.set_sensitive(state)
        self.radio_tracker_jellyfin.set_sensitive(state)

        if state:
            self._set_tracker_radio_buttons()
        else:
            self._enable_local(state)
            self._enable_plex(state)
            self._enable_jellyfin(state)

        self.checkbox_tracker_update_close.set_sensitive(state)
        self.checkbox_tracker_update_prompt.set_sensitive(state)
        self.checkbox_tracker_not_found_prompt.set_sensitive(state)

    @Gtk.Template.Callback()
    def _set_tracker_radio_buttons(self, radio_button=None):
        if self.radio_tracker_local.get_active() or self.radio_tracker_mpris.get_active():
            self._enable_local(True)
            self._enable_plex(False)
            self._enable_jellyfin(False)
        else:
            self._enable_local(False)
            self._enable_plex(True)
            self._enable_jellyfin(True)

    def _enable_local(self, enable):
        self.entry_player_process.set_sensitive(enable)
        self.btn_file_chooser_executable.set_sensitive(enable)
        self.checkbox_library_startup.set_sensitive(enable)
        self.checkbox_library_entire_list.set_sensitive(enable)
        self.checkbox_library_full_path.set_sensitive(enable)

    def _enable_plex(self, enable):
        self.entry_plex_host.set_sensitive(enable)
        self.spin_plex_port.set_sensitive(enable)
        self.entry_plex_username.set_sensitive(enable)
        self.entry_plex_password.set_sensitive(enable)
        self.checkbox_plex_obey_wait.set_sensitive(enable)
        self.checkbox_plex_ssl.set_sensitive(enable)

    def _enable_jellyfin(self, enable):
        self.entry_jellyfin_host.set_sensitive(enable)
        self.spin_jellyfin_port.set_sensitive(enable)
        self.entry_jellyfin_username.set_sensitive(enable)
        self.entry_jellyfin_api_key.set_sensitive(enable)

    def _load_directories(self, paths):
        if isinstance(paths, str):
            paths = [paths]

        for path in paths:
            self._add_row_listbox_directory(path)

    def _add_row_listbox_directory(self, path):
        row = DirectoryRow(path)
        self.listbox_directories.add(row)

    @Gtk.Template.Callback()
    def _on_btn_add_directory_clicked(self, btn):
        chooser_dialog = Gtk.FileChooserDialog('Select a directory',
                                               self.get_parent_window(),
                                               Gtk.FileChooserAction.OPEN,
                                               ("_Cancel", Gtk.ResponseType.CANCEL,
                                                "_Open", Gtk.ResponseType.OK))
        chooser_dialog.set_default_response(Gtk.ResponseType.OK)
        chooser_dialog.set_action(Gtk.FileChooserAction.SELECT_FOLDER)

        response = chooser_dialog.run()
        if response == Gtk.ResponseType.OK:
            self._add_row_listbox_directory(chooser_dialog.get_filename())
        chooser_dialog.destroy()

    def save_config(self):
        """Engine Configuration"""
        self.engine.set_config(
            'player', self.btn_file_chooser_executable.get_filename() or '')
        self.engine.set_config(
            'tracker_process', self.entry_player_process.get_text())
        self.engine.set_config('library_autoscan',
                               self.checkbox_library_startup.get_active())
        self.engine.set_config('scan_whole_list',
                               self.checkbox_library_entire_list.get_active())
        self.engine.set_config('library_full_path',
                               self.checkbox_library_full_path.get_active())
        self.engine.set_config('plex_host', self.entry_plex_host.get_text())
        self.engine.set_config('plex_port', str(
            int(self.spin_plex_port.get_value())))
        self.engine.set_config('plex_ssl',
                               self.checkbox_plex_ssl.get_active())
        self.engine.set_config('plex_obey_update_wait_s',
                               self.checkbox_plex_obey_wait.get_active())
        self.engine.set_config(
            'plex_user', self.entry_plex_username.get_text())
        self.engine.set_config(
            'plex_passwd', self.entry_plex_password.get_text())
        self.engine.set_config(
            'jellyfin_host', self.entry_jellyfin_host.get_text())
        self.engine.set_config('jellyfin_port', str(
            int(self.spin_jellyfin_port.get_value())))
        self.engine.set_config(
            'jellyfin_user', self.entry_jellyfin_username.get_text())
        self.engine.set_config(
            'jellyfin_api_key', self.entry_jellyfin_api_key.get_text())
        self.engine.set_config(
            'tracker_enabled', self.switch_tracker.get_active())
        self.engine.set_config(
            'autosend_at_exit', self.checkbox_upload_exit.get_active())
        self.engine.set_config('tracker_update_wait_s',
                               self.spin_tracker_update_wait.get_value_as_int())
        self.engine.set_config('tracker_update_close',
                               self.checkbox_tracker_update_close.get_active())
        self.engine.set_config('tracker_update_prompt',
                               self.checkbox_tracker_update_prompt.get_active())
        self.engine.set_config('tracker_not_found_prompt',
                               self.checkbox_tracker_not_found_prompt.get_active())

        self.engine.set_config(
            'searchdir', [row.directory for row in self.listbox_directories.get_children()])

        # Tracker type
        if self.radio_tracker_local.get_active():
            self.engine.set_config('tracker_type', 'local')
        elif self.radio_tracker_mpris.get_active():
            self.engine.set_config('tracker_type', 'mpris')
        elif self.radio_tracker_plex.get_active():
            self.engine.set_config('tracker_type', 'plex')
        elif self.radio_tracker_jellyfin.get_active():
            self.engine.set_config('tracker_type', 'jellyfin')

        # Auto-retrieve
        if self.radiobutton_download_always.get_active():
            self.engine.set_config('autoretrieve', 'always')
        elif self.radiobutton_download_days.get_active():
            self.engine.set_config('autoretrieve', 'days')
        else:
            self.engine.set_config('autoretrieve', 'off')

        # Auto-send
        if self.radiobutton_upload_always.get_active():
            self.engine.set_config('autosend', 'always')
        elif self.radiobutton_upload_minutes.get_active():
            self.engine.set_config('autosend', 'minutes')
        elif self.radiobutton_upload_size.get_active():
            self.engine.set_config('autosend', 'size')
        else:
            self.engine.set_config('autosend', 'off')

        self.engine.set_config(
            'autoretrieve_days', self.spinbutton_download_days.get_value_as_int())
        self.engine.set_config(
            'autosend_minutes', self.spinbutton_upload_minutes.get_value_as_int())
        self.engine.set_config(
            'autosend_size', self.spinbutton_upload_size.get_value_as_int())

        self.engine.set_config('auto_status_change',
                               self.checkbox_auto_status_change.get_active())
        self.engine.set_config('auto_status_change_if_scored',
                               self.checkbox_auto_status_change_if_scored.get_active())
        self.engine.set_config(
            'auto_date_change', self.checkbox_auto_date_change.get_active())
        self.engine.save_config()

        """GTK Interface configuration"""
        self.config['show_tray'] = self.checkbox_show_tray.get_active()

        if self.checkbox_show_tray.get_active():
            self.config['close_to_tray'] = self.checkbox_close_to_tray.get_active()
            self.config['start_in_tray'] = self.checkbox_start_in_tray.get_active()
            self.config['tray_api_icon'] = self.checkbox_tray_api_icon.get_active()
        else:
            self.config['close_to_tray'] = False
            self.config['start_in_tray'] = False
            self.config['tray_api_icon'] = False

        self.config['remember_geometry'] = self.checkbox_remember_geometry.get_active()
        self.config['episodebar_style'] = int(
            not self.checkbox_classic_progress.get_active())

        """Update Colors"""
        self.config['colors'] = {key: reprColor(
            col.get_color()) for key, col in self._color_buttons.items()}

        utils.save_config(self.config, self.configfile)


class DirectoryRow(Gtk.ListBoxRow):
    def __init__(self, directory):
        Gtk.ListBoxRow.__init__(self)

        self.directory = directory

        label = Gtk.Label(directory)
        label.set_ellipsize(Pango.EllipsizeMode.MIDDLE)

        image_button = Gtk.Image.new_from_icon_name('window-close-symbolic', 1)
        button_remove = Gtk.Button()
        button_remove.set_image(image_button)
        button_remove.connect('clicked', self._on_button_remove_click)

        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=9)
        box.pack_start(label, False, False, 0)
        box.pack_end(button_remove, False, False, 0)

        self.set_activatable(False)
        self.set_margin_bottom(5)
        self.set_margin_start(16)
        self.set_margin_end(16)
        self.set_margin_top(5)

        self.add(box)
        self.show_all()

    def _on_button_remove_click(self, btn):
        self.destroy()



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from gi import require_version
from gi.repository import GObject, Gdk, Gtk

from trackma import utils

require_version('Gdk', '3.0')


class TrackmaStatusIcon(Gtk.StatusIcon):
    __gtype_name__ = 'TrackmaStatusIcon'

    __gsignals__ = {
        'hide-clicked': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'about-clicked': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'quit-clicked': (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self):
        Gtk.StatusIcon.__init__(self)
        self.set_from_file(utils.DATADIR + '/icon.png')
        self.set_tooltip_text('Trackma GTK')
        self.connect('activate', self._tray_status_event)
        self.connect('popup-menu', self._tray_status_menu_event)

    def _tray_status_event(self, _widget):
        self.emit('hide-clicked')

    def _tray_status_menu_event(self, _icon, button, time):
        # Called when the tray icon is right-clicked
        menu = Gtk.Menu()
        mb_show = Gtk.MenuItem("Show/Hide")
        mb_about = Gtk.ImageMenuItem(
            'About', Gtk.Image.new_from_icon_name("help-about", Gtk.IconSize.MENU))
        mb_quit = Gtk.ImageMenuItem('Quit', Gtk.Image.new_from_icon_name(
            "application-exit", Gtk.IconSize.MENU))

        mb_show.connect("activate", self._tray_status_event)
        mb_about.connect("activate", self._on_mb_about)
        mb_quit.connect("activate", self._on_mb_quit)

        menu.append(mb_show)
        menu.append(mb_about)
        menu.append(Gtk.SeparatorMenuItem())
        menu.append(mb_quit)
        menu.show_all()

        menu.popup(None, None, None, self._pos, button, time)

    def _on_mb_about(self, menu_item):
        self.emit('about-clicked')

    def _on_mb_quit(self, menu_item):
        self.emit('quit-clicked')

    @staticmethod
    def _pos(menu, icon):
        return Gtk.StatusIcon.position_menu(menu, icon)

    @staticmethod
    def is_tray_available():
        # Icon tray isn't available in Wayland
        return not Gdk.Display.get_default().get_name().lower().startswith('wayland')



from PyQt6 import QtCore, QtGui
from PyQt6.QtWidgets import QDoubleSpinBox, QStyle, QStyleOptionProgressBar, QStyledItemDelegate

from trackma.ui.qt.util import getColor

MARGIN = 5
PADDING = 5
WIDTH = 450
MIN_HEIGHT = 200
COLUMN_A = 100
COLUMN_B = 290


class AddListDelegate(QStyledItemDelegate):
    """ This is the delegate that handles the rendering of cards
    in the List View of the Add show dialog. """

    def __init__(self, parent=None):
        self.results = None

        self.font = QtGui.QFont()

        fm = QtGui.QFontMetrics(self.font)
        self.fh = fm.height()

        # Get theme colors
        palette = QtGui.QPalette()
        self.alternatebasecolor = palette.color(palette.ColorRole.AlternateBase)
        self.windowtextcolor = palette.color(palette.ColorRole.WindowText)
        self.windowcolor = palette.color(palette.ColorRole.Window)

        super().__init__(parent)

    def _get_extra(self, extra, key):
        for k, v in extra:
            if k == key:
                return v

    def paint(self, painter, option, index):
        outerRect = option.rect - \
            QtCore.QMargins(MARGIN, MARGIN, MARGIN, MARGIN)

        data = index.data()
        thumb = index.data(QtCore.Qt.ItemDataRole.DecorationRole)

        painter.save()

        color = index.data(QtCore.Qt.ItemDataRole.BackgroundRole)

        # Draw background box
        painter.setPen(QtGui.QPen(self.alternatebasecolor))
        painter.setBrush(QtGui.QBrush(self.windowcolor))
        painter.drawRect(outerRect)

        # Prepare to draw inside
        baseRect = outerRect - \
            QtCore.QMargins(PADDING, PADDING, PADDING, PADDING)
        painter.setPen(QtCore.Qt.PenStyle.NoPen)

        # Draw thumbnail (if any)
        if thumb:
            painter.drawImage(baseRect.topLeft(), thumb)

        # Create text QRect and draw the title background
        textRect = baseRect.adjusted(COLUMN_A+5, 0, 0, 0)
        textRect.setHeight(self.fh + 5)
        painter.setBrush(QtGui.QBrush(color))
        painter.drawRect(textRect)

        # Set our font to bold
        bfont = QtGui.QFont(self.font)
        bfont.setWeight(QtGui.QFont.Weight.Bold)

        painter.setFont(bfont)
        painter.setPen(QtGui.QPen(QtGui.QColor(10, 10, 10)))

        # Make some padding
        textRect -= QtCore.QMargins(5, 0, 5, 0)

        # Draw title
        painter.drawText(textRect, QtCore.Qt.AlignmentFlag.AlignVCenter, data['title'])

        painter.setPen(QtGui.QPen(self.windowtextcolor))

        # Draw the details
        textRect.setHeight(self.fh)
        dataRect = textRect.adjusted(75, 0, 0, 0)

        textRect.translate(0, self.fh + 10)
        painter.drawText(textRect, QtCore.Qt.AlignmentFlag.AlignTop, "Date")
        textRect.translate(0, self.fh + 5)
        painter.drawText(textRect, QtCore.Qt.AlignmentFlag.AlignTop, "Episodes")

        # Draw data
        painter.setFont(self.font)

        # Dates
        if data.get('start_date'):
            d_from = data['start_date'].strftime('%d/%m/%y')
        else:
            d_from = '?'
        if data.get('end_date'):
            d_end = data['end_date'].strftime('%d/%m/%y')
        else:
            d_end = '?'

        dataRect.translate(0, self.fh + 10)
        painter.drawText(dataRect, QtCore.Qt.AlignmentFlag.AlignTop,
                         "{} to {}".format(d_from, d_end))
        dataRect.translate(0, self.fh + 5)
        painter.drawText(dataRect, QtCore.Qt.AlignmentFlag.AlignTop,
                         str(data.get('total') or '?'))

        # Draw synopsis
        textRect.translate(0, self.fh + 5)
        textRect.setBottomRight(baseRect.bottomRight())

        if 'extra' in data:
            painter.drawText(textRect, QtCore.Qt.AlignmentFlag.AlignTop | QtCore.Qt.TextFlag.TextWordWrap, self._get_extra(
                data['extra'], 'Synopsis'))

        # Draw select box
        if option.state & QStyle.StateFlag.State_Selected:
            painter.setCompositionMode(QtGui.QPainter.CompositionMode.CompositionMode_Overlay)
            # painter.setOpacity(0.5)
            painter.fillRect(outerRect, option.palette.highlight())

        painter.restore()

    def sizeHint(self, option, index):
        return QtCore.QSize(WIDTH, min(MIN_HEIGHT, self.fh*10+15))


class ShowsTableDelegate(QStyledItemDelegate):
    """
    Custom delegate that shows a custom progress bar
    for detailed information about episodes, and editing
    the progress and score.
    """
    # Enum BarStyle
    BarStyleBasic = 0   # Basic native ProgressBar appearance
    BarStyle04 = 1      # Rectangular dual bar of Trackma v0.4
    BarStyleHybrid = 2  # Native ProgressBar with v0.4 library subbar overlaid

    _subvalue = -1
    _episodes = []
    _subheight = 5
    _bar_style = BarStyle04
    _show_text = False

    def __init__(self, parent, palette=None):
        self.colors = palette

        super().__init__(parent)

    def paint(self, painter, option, index):
        if index.column() == 4:
            rect = option.rect
            data = index.model().data(index)

            if not data:
                return

            (value, maximum, subvalue, episodes) = data
            m = index.model().sourceModel()

            painter.save()

            if self._bar_style is self.BarStyleBasic:
                prog_options = QStyleOptionProgressBar()
                prog_options.maximum = maximum
                prog_options.progress = value
                prog_options.rect = rect
                prog_options.text = '%d%%' % (value*100/maximum)
                prog_options.textVisible = self._show_text
                option.widget.style().drawControl(QStyle.ControlElement.CE_ProgressBar, prog_options, painter)

            elif self._bar_style is self.BarStyle04:
                painter.setBrush(getColor(self.colors['progress_bg']))
                painter.setPen(QtCore.Qt.GlobalColor.transparent)
                painter.drawRect(rect)
                self.paintSubValue(painter, rect, subvalue, maximum)
                if value > 0:
                    if value >= maximum:
                        painter.setBrush(
                            getColor(self.colors['progress_complete']))
                        mid = rect.width()
                    else:
                        painter.setBrush(getColor(self.colors['progress_fg']))
                        mid = int(rect.width() / float(maximum) * value)
                    progressRect = QtCore.QRect(
                        rect.x(), rect.y(), mid, rect.height())
                    painter.drawRect(progressRect)
                self.paintEpisodes(painter, rect, episodes, maximum)

            elif self._bar_style is self.BarStyleHybrid:
                painter.setCompositionMode(
                    QtGui.QPainter.CompositionMode.CompositionMode_Source)
                painter.fillRect(rect, QtCore.Qt.GlobalColor.transparent)
                painter.setCompositionMode(
                    QtGui.QPainter.CompositionMode.CompositionMode_SourceOver)
                prog_options = QStyleOptionProgressBar()
                prog_options.maximum = maximum
                prog_options.progress = value
                prog_options.rect = rect
                prog_options.text = '%d%%' % (value*100/maximum)
                option.widget.style().drawControl(QStyle.ControlElement.CE_ProgressBar, prog_options, painter)
                painter.setCompositionMode(
                    QtGui.QPainter.CompositionMode.CompositionMode_SourceAtop)
                painter.setPen(QtCore.Qt.GlobalColor.transparent)
                self.paintSubValue(painter, rect, subvalue, maximum)
                self.paintEpisodes(painter, rect, episodes, maximum)
                painter.setCompositionMode(
                    QtGui.QPainter.CompositionMode.CompositionMode_SourceOver)
                if self._show_text:
                    option.widget.style().drawControl(QStyle.ControlElement.CE_ProgressBarLabel, prog_options, painter)

            painter.restore()
        else:
            super().paint(painter, option, index)

    def paintSubValue(self, painter, rect, subvalue, maximum):
        if subvalue and maximum and subvalue <= maximum:
            painter.setBrush(getColor(self.colors['progress_sub_bg']))
            mid = int(rect.width() / float(maximum) * subvalue)
            progressRect = QtCore.QRect(
                rect.x(),
                rect.y()+rect.height()-self._subheight,
                mid,
                rect.height()-(rect.height()-self._subheight)
            )
            painter.drawRect(progressRect)

    def paintEpisodes(self, painter, rect, episodes, maximum):
        if episodes:
            for episode in episodes:
                painter.setBrush(getColor(self.colors['progress_sub_fg']))
                if episode <= maximum:
                    start = int(rect.width() / float(maximum) * (episode - 1))
                    finish = int(rect.width() / float(maximum) * episode)
                    progressRect = QtCore.QRect(
                        rect.x()+start,
                        rect.y()+rect.height()-self._subheight,
                        finish-start,
                        rect.height()-(rect.height()-self._subheight)
                    )
                    painter.drawRect(progressRect)

    def setBarStyle(self, style, show_text):
        self._bar_style = style
        self._show_text = show_text

    def sizeHint(self, option, index):
        return QtCore.QSize(option.rect.width(), QtGui.QFontMetrics(option.font).height() + 2)

    def createEditor(self, parent, option, index):
        editor = QDoubleSpinBox(parent)
        editor.setFrame(False)

        return editor

    def setEditorData(self, editor, index):
        (value, maximum, decimals, step) = index.model().data(
            index, QtCore.Qt.ItemDataRole.EditRole)

        editor.setMaximum(maximum or 999)
        editor.setDecimals(decimals or 0)
        editor.setSingleStep(step or 1)

        if value:
            editor.setValue(value)

    def setModelData(self, editor, model, index):
        editor.interpretText()
        old_value = index.model().data(index, QtCore.Qt.ItemDataRole.EditRole)[0]
        new_value = editor.value()

        if new_value != old_value:
            model.setData(index, new_value, QtCore.Qt.ItemDataRole.EditRole)



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import subprocess
import sys
import threading

from gi.repository import GLib, Gdk, Gio, Gtk

from trackma import messenger
from trackma import utils
from trackma.accounts import AccountManager
from trackma.engine import Engine
from trackma.ui.gtk import gtk_dir
from trackma.ui.gtk.accountswindow import AccountsWindow
from trackma.ui.gtk.mainview import MainView
from trackma.ui.gtk.searchwindow import SearchWindow
from trackma.ui.gtk.settingswindow import SettingsWindow
from trackma.ui.gtk.showeventtype import ShowEventType
from trackma.ui.gtk.showinfowindow import ShowInfoWindow
from trackma.ui.gtk.statusicon import TrackmaStatusIcon


@Gtk.Template.from_file(os.path.join(gtk_dir, 'data/window.ui'))
class TrackmaWindow(Gtk.ApplicationWindow):
    __gtype_name__ = 'TrackmaWindow'

    btn_appmenu = Gtk.Template.Child()
    btn_mediatype = Gtk.Template.Child()
    header_bar = Gtk.Template.Child()

    def __init__(self, app, debug=False):
        Gtk.ApplicationWindow.__init__(self, application=app)
        self.init_template()

        self._debug = debug
        self._configfile = utils.to_config_path('ui-Gtk.json')
        self._config = utils.parse_config(self._configfile, utils.gtk_defaults)

        self.statusicon = None
        self._main_view = None
        self._modals = []

        self._account = None
        self._engine = None
        self.close_thread = None
        self.hidden = False

        self._init_widgets()

    def init_account_selection(self):
        manager = AccountManager()

        # Use the remembered account if there's one
        if manager.get_default():
            self._create_engine(manager.get_default())
        else:
            self._show_accounts(switch=False)

    def _init_widgets(self):
        Gtk.Window.set_default_icon_from_file(utils.DATADIR + '/icon.png')
        self.set_position(Gtk.WindowPosition.CENTER)
        self.set_title('Trackma')

        if self._config['remember_geometry']:
            self.resize(self._config['last_width'],
                        self._config['last_height'])

        if not self._main_view:
            self._main_view = MainView(self._config)
            self._main_view.connect('error', self._on_main_view_error)
            self._main_view.connect(
                'success', lambda x: self._set_buttons_sensitive(True))
            self._main_view.connect(
                'error-fatal', self._on_main_view_error_fatal)
            self._main_view.connect('show-action', self._on_show_action)
            self.add(self._main_view)

        self.connect('delete_event', self._on_delete_event)

        builder = Gtk.Builder.new_from_file(
            os.path.join(gtk_dir, 'data/shortcuts.ui'))
        help_overlay = builder.get_object('shortcuts-window')
        self.set_help_overlay(help_overlay)

        # Status icon
        if TrackmaStatusIcon.is_tray_available():
            self.statusicon = TrackmaStatusIcon()
            self.statusicon.connect('hide-clicked', self._on_tray_hide_clicked)
            self.statusicon.connect(
                'about-clicked', self._on_tray_about_clicked)
            self.statusicon.connect('quit-clicked', self._on_tray_quit_clicked)

            if self._config['show_tray']:
                self.statusicon.set_visible(True)
            else:
                self.statusicon.set_visible(False)

        # Don't show the main window if start in tray option is set
        if self.statusicon and self._config['show_tray'] and self._config['start_in_tray']:
            self.hidden = True
        else:
            self.present()

    def _on_tray_hide_clicked(self, status_icon):
        self._destroy_modals()

        if self.hidden:
            self.deiconify()
            self.present()

            if not self._engine:
                self._show_accounts(switch=False)
        else:
            self.hide()

        self.hidden = not self.hidden

    def _destroy_modals(self):
        self.get_help_overlay().hide()

        for modal_window in self._modals:
            modal_window.destroy()

        self._modals = []

    def _on_tray_about_clicked(self, status_icon):
        self._on_about(None, None)

    def _on_tray_quit_clicked(self, status_icon):
        self._quit()

    def _on_delete_event(self, widget, event, data=None):
        if self.statusicon and self.statusicon.get_visible() and self._config['close_to_tray']:
            self.hidden = True
            self.hide()
        else:
            self._quit()
        return True

    def _create_engine(self, account):
        self._engine = Engine(account, self._message_handler)

        self._main_view.load_engine_account(self._engine, account)
        self._set_actions()
        self._set_mediatypes_menu()
        self._update_widgets(account)
        self._set_buttons_sensitive(True)

    def _set_actions(self):
        builder = Gtk.Builder.new_from_file(
            os.path.join(gtk_dir, 'data/app-menu.ui'))
        settings = Gtk.Settings.get_default()
        if not settings.get_property("gtk-shell-shows-menubar"):
            self.btn_appmenu.set_menu_model(builder.get_object('app-menu'))
        else:
            self.get_application().set_menubar(builder.get_object('menu-bar'))
            self.btn_appmenu.set_property('visible', False)

        def add_action(name, callback):
            action = Gio.SimpleAction.new(name, None)
            action.connect('activate', callback)
            self.add_action(action)

        add_action('search', self._on_search)
        add_action('synchronize', self._on_synchronize)
        add_action('upload', self._on_upload)
        add_action('download', self._on_download)
        add_action('scanfiles', self._on_scanfiles)
        add_action('accounts', self._on_accounts)
        add_action('preferences', self._on_preferences)
        add_action('about', self._on_about)

        add_action('play_next', self._on_action_play_next)
        add_action('play_random', self._on_action_play_random)
        add_action('episode_add', self._on_action_episode_add)
        add_action('episode_remove', self._on_action_episode_remove)
        add_action('delete', self._on_action_delete)
        add_action('copy', self._on_action_copy)

    def _set_mediatypes_action(self):
        action_name = 'change-mediatype'
        if self.has_action(action_name):
            self.remove_action(action_name)

        state = GLib.Variant.new_string(self._engine.api_info['mediatype'])
        action = Gio.SimpleAction.new_stateful(action_name,
                                               state.get_type(),
                                               state)
        action.connect('change-state', self._on_change_mediatype)
        self.add_action(action)

    def _set_mediatypes_menu(self):
        self._set_mediatypes_action()
        menu = Gio.Menu()

        for mediatype in self._engine.api_info['supported_mediatypes']:
            variant = GLib.Variant.new_string(mediatype)
            menu_item = Gio.MenuItem()
            menu_item.set_label(mediatype)
            menu_item.set_action_and_target_value(
                'win.change-mediatype', variant)
            menu.append_item(menu_item)

        self.btn_mediatype.set_menu_model(menu)

        if len(self._engine.api_info['supported_mediatypes']) <= 1:
            self.btn_mediatype.hide()

    def _update_widgets(self, account):
        current_api = utils.available_libs[account['api']]
        api_iconpath = 1
        api_iconfile = current_api[api_iconpath]

        self.header_bar.set_subtitle(self._engine.api_info['name'] + " (" +
                                     self._engine.api_info['mediatype'] + ")")

        if self.statusicon and self._config['tray_api_icon']:
            self.statusicon.set_from_file(api_iconfile)

    def _on_change_mediatype(self, action, value):
        action.set_state(value)
        mediatype = value.get_string()
        self._set_buttons_sensitive(False)
        self._main_view.load_account_mediatype(
            None, mediatype, self.header_bar)

    def _on_search(self, action, param):
        current_status = self._main_view.get_current_status()
        win = SearchWindow(
            self._engine, self._config['colors'], current_status, transient_for=self)
        win.connect('search-error', self._on_search_error)
        win.connect('destroy', self._on_modal_destroy)
        win.present()
        self._modals.append(win)

    def _on_search_error(self, search_window, error_msg):
        print(error_msg)

    def _on_synchronize(self, action, param):
        threading.Thread(target=self._synchronization_task,
                         args=(True, True)).start()

    def _on_upload(self, action, param):
        threading.Thread(target=self._synchronization_task,
                         args=(True, False)).start()

    def _on_download(self, action, param):
        def _download_lists():
            threading.Thread(target=self._synchronization_task,
                             args=(False, True)).start()

        def _on_download_response(_dialog, response):
            _dialog.destroy()

            if response == Gtk.ResponseType.YES:
                _download_lists()

        queue = self._engine.get_queue()
        if queue:
            dialog = Gtk.MessageDialog(self,
                                       Gtk.DialogFlags.MODAL,
                                       Gtk.MessageType.QUESTION,
                                       Gtk.ButtonsType.YES_NO,
                                       "There are %d queued changes in your list. If you retrieve the remote list now you will lose your queued changes. Are you sure you want to continue?" % len(queue))
            dialog.show_all()
            dialog.connect("response", _on_download_response)
        else:
            # If the user doesn't have any queued changes
            # just go ahead
            _download_lists()

    def _synchronization_task(self, send, retrieve):
        self._set_buttons_sensitive_idle(False)

        try:
            if send:
                self._engine.list_upload()
            if retrieve:
                self._engine.list_download()

            # GLib.idle_add(self._set_score_ranges)
            GLib.idle_add(self._main_view.populate_all_pages)
        except utils.TrackmaError as e:
            self._error_dialog_idle(e)
        except utils.TrackmaFatal as e:
            self._show_accounts_idle(switch=False, forget=True)
            self._error_dialog_idle("Fatal engine error: %s" % e)
            return

        self._main_view.set_status_idle("Ready.")
        self._set_buttons_sensitive_idle(True)

    def _on_scanfiles(self, action, param):
        threading.Thread(target=self._scanfiles_task).start()

    def _scanfiles_task(self):
        self._set_buttons_sensitive_idle(False)
        try:
            self._engine.scan_library(rescan=True)
        except utils.TrackmaError as e:
            self._error_dialog_idle(e)

        GLib.idle_add(self._main_view.populate_all_pages)

        self._main_view.set_status_idle("Ready.")
        self._set_buttons_sensitive_idle(True)

    def _on_accounts(self, action, param):
        self._show_accounts()

    def _show_accounts_idle(self, switch=True, forget=False):
        GLib.idle_add(self._show_accounts, switch, forget)

    def _show_accounts(self, switch=True, forget=False):
        manager = AccountManager()

        if forget:
            manager.set_default(None)

        accountsel = AccountsWindow(manager, transient_for=self)
        accountsel.connect('account-open', self._on_account_open)
        accountsel.connect('account-cancel', self._on_account_cancel, switch)
        accountsel.connect('destroy', self._on_modal_destroy)
        accountsel.present()
        self._modals.append(accountsel)

    def _on_account_open(self, accounts_window, account_num, remember):
        manager = AccountManager()
        account = manager.get_account(account_num)

        if remember:
            manager.set_default(account_num)
        else:
            manager.set_default(None)

        # Reload the engine if already started,
        # start it otherwise
        self._set_buttons_sensitive(False)
        if self._engine and self._engine.loaded:
            self._main_view.load_account_mediatype(account, None, None)
        else:
            self._create_engine(account)

    def _on_account_cancel(self, _accounts_window, switch):
        manager = AccountManager()

        if not switch or not manager.get_accounts():
            self._quit()

    def _on_preferences(self, _action, _param):
        win = SettingsWindow(self._engine, self._config,
                             self._configfile, transient_for=self)
        win.connect('destroy', self._on_modal_destroy)
        win.present()
        self._modals.append(win)

    def _on_about(self, _action, _param):
        about = Gtk.AboutDialog(parent=self)
        about.set_modal(True)
        about.set_transient_for(self)
        about.set_program_name("Trackma GTK")
        about.set_version(utils.VERSION)
        about.set_license_type(Gtk.License.GPL_3_0_ONLY)
        about.set_comments(
            "Trackma is an open source client for media tracking websites.\nThanks to all contributors.")
        about.set_website("https://github.com/z411/trackma")
        about.set_copyright("© z411, et al.")
        about.set_authors(["See AUTHORS file"])
        about.set_artists(["shuuichi"])
        about.connect('destroy', self._on_modal_destroy)
        about.connect('response', lambda dialog, response: dialog.destroy())
        about.present()
        self._modals.append(about)

    def _on_modal_destroy(self, modal_window):
        self._modals.remove(modal_window)

    def _quit(self):
        if self._config['remember_geometry']:
            self._store_geometry()

        if not self._engine:
            self.get_application().quit()
            return

        if self.close_thread is None:
            self._set_buttons_sensitive_idle(False)
            self.close_thread = threading.Thread(target=self._unload_task)
            self.close_thread.start()

    def _unload_task(self):
        self._engine.unload()
        GLib.idle_add(self.get_application().quit)

    def _store_geometry(self):
        (width, height) = self.get_size()
        self._config['last_width'] = width
        self._config['last_height'] = height
        utils.save_config(self._config, self._configfile)

    def _message_handler(self, classname, msgtype, msg):
        # Thread safe
        # print("%s: %s" % (classname, msg))
        if msgtype == messenger.TYPE_WARN:
            self._main_view.set_status_idle(
                "%s warning: %s" % (classname, msg))
        elif msgtype != messenger.TYPE_DEBUG:
            self._main_view.set_status_idle("%s: %s" % (classname, msg))
        elif self._debug:
            print('[D] {}: {}'.format(classname, msg))

    def _on_main_view_error(self, main_view, error_msg):
        self._error_dialog_idle(error_msg)

    def _on_main_view_error_fatal(self, main_view, error_msg):
        self._show_accounts_idle(switch=False, forget=True)
        self._error_dialog_idle(error_msg)

    def _error_dialog_idle(self, msg, icon=Gtk.MessageType.ERROR):
        # Thread safe
        GLib.idle_add(self._error_dialog, msg, icon)

    def _error_dialog(self, msg, icon=Gtk.MessageType.ERROR):
        def error_dialog_response(widget, response_id):
            widget.destroy()

        dialog = Gtk.MessageDialog(self,
                                   Gtk.DialogFlags.MODAL,
                                   icon,
                                   Gtk.ButtonsType.OK,
                                   str(msg))
        dialog.show_all()
        dialog.connect("response", error_dialog_response)
        print('Error: {}'.format(msg))

    def _on_action_play_next(self, action, param):
        selected_show = self._main_view.get_selected_show()

        if selected_show:
            self._play_next(selected_show)

    def _on_action_play_random(self, action, param):
        self._play_random()

    def _on_action_episode_add(self, action, param):
        selected_show = self._main_view.get_selected_show()

        if selected_show:
            self._episode_add(selected_show)

    def _on_action_episode_remove(self, action, param):
        selected_show = self._main_view.get_selected_show()

        if selected_show:
            self._episode_remove(selected_show)

    def _on_action_delete(self, action, param):
        selected_show = self._main_view.get_selected_show()

        if selected_show:
            self._remove_show(selected_show)

    def _on_action_copy(self, action, param):
        selected_show = self._main_view.get_selected_show()

        if selected_show:
            self._copy_title(selected_show)

    def _on_show_action(self, main_view, event_type, data):
        if event_type == ShowEventType.PLAY_NEXT:
            self._play_next(*data)
        elif event_type == ShowEventType.PLAY_EPISODE:
            self._play_episode(*data)
        elif event_type == ShowEventType.EPISODE_REMOVE:
            self._episode_remove(*data)
        elif event_type == ShowEventType.EPISODE_SET:
            self._episode_set(*data)
        elif event_type == ShowEventType.EPISODE_ADD:
            self._episode_add(*data)
        elif event_type == ShowEventType.SET_SCORE:
            self._set_score(*data)
        elif event_type == ShowEventType.SET_STATUS:
            self._set_status(*data)
        elif event_type == ShowEventType.DETAILS:
            self._open_details(*data)
        elif event_type == ShowEventType.OPEN_WEBSITE:
            self._open_website(*data)
        elif event_type == ShowEventType.OPEN_FOLDER:
            self._open_folder(*data)
        elif event_type == ShowEventType.COPY_TITLE:
            self._copy_title(*data)
        elif event_type == ShowEventType.CHANGE_ALTERNATIVE_TITLE:
            self._change_alternative_title(*data)
        elif event_type == ShowEventType.REMOVE:
            self._remove_show(*data)

    def _play_next(self, show_id):
        show = self._engine.get_show_info(show_id)
        try:
            args = self._engine.play_episode(show)
            utils.spawn_process(args)
        except utils.TrackmaError as e:
            self._error_dialog(e)

    def _play_episode(self, show_id, episode):
        show = self._engine.get_show_info(show_id)
        try:
            if not episode:
                episode = self.show_ep_num.get_value_as_int()
            args = self._engine.play_episode(show, episode)
            utils.spawn_process(args)
        except utils.TrackmaError as e:
            self._error_dialog(e)

    def _play_random(self):
        try:
            args = self._engine.play_random()
            utils.spawn_process(args)
        except utils.TrackmaError as e:
            self._error_dialog(e)

    def _episode_add(self, show_id):
        show = self._engine.get_show_info(show_id)
        self._episode_set(show_id, show['my_progress'] + 1)

    def _episode_remove(self, show_id):
        show = self._engine.get_show_info(show_id)
        self._episode_set(show_id, show['my_progress'] - 1)

    def _episode_set(self, show_id, episode):
        try:
            self._engine.set_episode(show_id, episode)
        except utils.TrackmaError as e:
            self._error_dialog(e)

    def _set_score(self, show_id, score):
        try:
            self._engine.set_score(show_id, score)
        except utils.TrackmaError as e:
            self._error_dialog(e)

    def _set_status(self, show_id, status):
        try:
            self._engine.set_status(show_id, status)
        except utils.TrackmaError as e:
            self._error_dialog(e)

    def _open_details(self, show_id):
        show = self._engine.get_show_info(show_id)
        win = ShowInfoWindow(self._engine, show, transient_for=self)
        win.connect('destroy', self._on_modal_destroy)
        win.present()
        self._modals.append(win)

    def _open_website(self, show_id):
        show = self._engine.get_show_info(show_id)
        if show['url']:
            Gtk.show_uri(None, show['url'], Gdk.CURRENT_TIME)

    def _open_folder(self, show_id):
        try:
            self._engine.open_show_folder(show_id)
        except utils.EngineError as e:
            self._error_dialog_idle(e.args[0])

    def _copy_title(self, show_id):
        show = self._engine.get_show_info(show_id)
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(show['title'], -1)

        self._main_view.set_status_idle('Title copied to clipboard.')

    def _change_alternative_title(self, show_id):
        show = self._engine.get_show_info(show_id)
        current_altname = self._engine.altname(show_id)

        def altname_response(entry, dialog, response):
            dialog.response(response)

        dialog = Gtk.MessageDialog(
            self,
            Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
            Gtk.MessageType.QUESTION,
            Gtk.ButtonsType.OK_CANCEL,
            None)
        dialog.set_markup('Set the <b>alternate title</b> for the show.')
        entry = Gtk.Entry()
        entry.set_text(current_altname)
        entry.connect("activate", altname_response,
                      dialog, Gtk.ResponseType.OK)
        hbox = Gtk.HBox()
        hbox.pack_start(Gtk.Label("Alternate Title:"), False, 5, 5)
        hbox.pack_end(entry, True, True, 0)
        dialog.format_secondary_markup(
            "Use this if the tracker is unable to find this show. Leave blank to disable.")
        dialog.vbox.pack_end(hbox, True, True, 0)
        dialog.show_all()
        retval = dialog.run()

        if retval == Gtk.ResponseType.OK:
            text = entry.get_text()
            self._engine.altname(show_id, text)
            self._main_view.change_show_title_idle(show, text)

        dialog.destroy()

    def _remove_show(self, show_id):
        try:
            show = self._engine.get_show_info(show_id)
            self._engine.delete_show(show)
        except utils.TrackmaError as e:
            self._error_dialog_idle(e)

    def _set_buttons_sensitive_idle(self, sensitive):
        GLib.idle_add(self._set_buttons_sensitive, sensitive)
        self._main_view.set_buttons_sensitive_idle(sensitive)

    def _set_buttons_sensitive(self, sensitive):
        actions_names = ['search',
                         'synchronize',
                         'upload',
                         'download',
                         'scanfiles',
                         'accounts',
                         'play_next',
                         'play_random',
                         'episode_add',
                         'episode_remove',
                         'delete',
                         'copy']

        for action_name in actions_names:
            action = self.lookup_action(action_name)

            if action is not None:
                action.set_enabled(sensitive)
